class ApiController < ActionController::API
  class RequiredFilterNotProvided < StandardError; end
  class RequiredCombinationFilterNotProvided < StandardError; end
  class UndefinedColumnOnFilter < StandardError; end
  class MetaQueryIsNotDefined < StandardError; end
  class UndefinedSelectedColumns < StandardError; end
  class EnumRuleBroken < StandardError; end
  class UndefinedOrderColumn < StandardError; end

  def collect_metrics
    start = Time.now
    yield
    Time.now - start
  end

  def return_data_root
    :root
  end

  def base_filters
    {
      limit: {
        meta: true,
        param: :limit,
        value: 1000
      },
      order: {
        meta: true,
        param: :order,
        value: :DESC,
        column: order_column
      },
      offset: {
        meta: true,
        param: :offset,
        value: 0,
      }
    }
  end

  def selected_columns
    # redefine in the api action with a list of strings with each selected column
    # def select
    #   %w[jobs.number side state substate process comment scrap scrap_reason time machines.sn]
    # end
    raise UndefinedSelectedColumns
  end

  def filters
    # redefine in the api action with object following this format
    # {
    #   time: {
    #     column: "time",
    #     required: true,
    #     range: {
    #       lower_bound: {
    #         param: :start,
    #       },
    #       upper_bound: {
    #         param: :end
    #       }
    #     }
    #   },
    #   sn: {
    #     column: "machine.sn",
    #     param: :sn,
    #     required_combination: [:job]
    #   },
    #   job: {
    #     column: "job.number",
    #     param: :job,
    #     required_combination: [:sn]
    #   }
    # }
    {}
  end

  # Sets the model that will be queried on. Joins are to be set here
  def set_model(model)
    @model = model
    @query = @model
    @null_filters = {}
    @metrics = {}
  end

  @use_archive_shard = false

  # Group a column together
  def group_by(attribute)
    @group_by = attribute
  end

  # Get grouped values to be aggregated under the aggregation_name
  def aggregate(name, columns)
    @aggregation_name = name
    @aggregate_columns = columns
  end

  def order_column
    raise UndefinedOrderColumn
  end

  def debug?
    params.include? "debug"
  end

  # Adds the meta query parameters, like :order or :offset,
  def query_meta_builder(filter)
    case filter[:param]
    when :order
      raise UndefinedColumnOnFilter unless filter[:column]
      @query = @query.order("#{filter[:column]} #{filter[:value]}")
    when :limit
      @query = @query.limit(filter[:value])
    when :offset
      @query = @query.offset(filter[:value])
    else
      raise MetaQueryIsNotDefined
    end
  end

  # Adds a range to the query. one or two bound can be set
  def query_range_builder(filter)
    column = filter[:column]
    lower_bound_value = filter[:range][:lower_bound][:value]
    upper_bound_value = filter[:range][:upper_bound][:value]
    if %w[time created_at updated_at].include? column
      @use_archive_shard = Time.parse(lower_bound_value) < (Time.now - 32.days)
    end
    @query = @query.where(@model.arel_table[column].gteq(lower_bound_value)) if lower_bound_value
    @query = @query.where(@model.arel_table[column].lteq(upper_bound_value)) if upper_bound_value
  end

  # Handles all equal filters
  def query_equal_builder(filter)
    column = filter[:column]
    # in is used as the equal operator since it handles both arrays an singular values
    @query = @query.where(find_arel(@model, column).in(filter[:value]))
  end

  # Some filters have combination requirement, at least one of them needs to be passed.
  def ensure_required_combination_filters
    filters_to_remove = []
    @filters.each do |id, filter|
      next unless filter[:required_combination]
      filter_ids = [id] + filter[:required_combination]
      if filter_ids.all? {|r_c_id| @null_filters.has_key?(r_c_id)}
        Rails::logger.error "Required Combination Filter not given."
        raise RequiredFilterNotProvided
      end
      filters_to_remove.push(id) if @null_filters.has_key?(id)
    end
    filters_to_remove.each { |id| @filters.delete(id) }
  end

  def ensure_enum_rules
    @filters.each do |column, filter|
      next if column == :time
      next unless @model.defined_enums
      next unless @model.defined_enums.has_key? column.to_s
      if filter[:value].is_a? Array
        raise EnumRuleBroken unless filter[:value].any? {|val|@model.defined_enums[column.to_s].has_key? val }
      else
        raise EnumRuleBroken unless @model.defined_enums[column.to_s].has_key? filter[:value]
      end
    end
  end

  # Base method that handles all the heavy lifting
  def do_query
    begin
      resp = {}
      dur = collect_metrics do
        @filters = filters
        @filters = @filters.merge(base_filters)

        build_filters

        ensure_required_combination_filters

        ensure_enum_rules

        build_query_from_filters

        calc_next_offset

        ActiveRecord::Base.connected_to(shard: @use_archive_shard ? :archive : :primary, role: :reading) do
          pluck_to_models
        end

        model_side_num_to_L_R
        group
        @metrics[:db] = @use_archive_shard ? "archive" : "primary"

        resp[:next_offset] = @next_offset
        resp[return_data_root] = @models
        resp[:model_count] = @models.count
      end
      @metrics[:request] = "The request took #{dur} seconds to be processed"
      resp[:metrics] = @metrics if debug?
      render json: resp, status: :ok
    rescue => e
      Rails::logger.warn "API ERROR | #{e}"
      render json: {error: e}, status: :unprocessable_entity
    end
  end

  # Handles setting the values in the the range filters
  def filter_range_builder(id, filter)
    [:lower_bound, :upper_bound].each do |bound|
      if filter[:required]
        unless params[filter[:range][bound][:param]]
          Rails::logger.error "Required filter #{filter[:range][bound][:param]} not given."
          raise RequiredFilterNotProvided
        end
        filter[:range][bound][:value] = params[filter[:range][bound][:param]]
      else
        if filter[:required_combination]
          @null_filters[id] = nil unless params[filter[:range][bound][:param]]
          filter[:range][bound][:value] = params[filter[:range][bound][:param]]
          next
        end

        if params[filter[:range][bound][:param]]
          filter[:range][bound][:value] = params[filter[:range][bound][:param]]
        else
          @filters_not_used.push(id)
        end
      end
    end
  end

  def build_query_from_filters
    @filters.each do |id, filter|
      if filter[:meta]
        next query_meta_builder(filter)
      end
      if filter[:range]
        next query_range_builder(filter)
      end
      query_equal_builder(filter)
    end
  end

  # Goes over the filters and sets all the values from the parameters.
  # Also removes unused filters.
  def build_filters
    @filters_not_used = []
    @filters.each do |id, filter|

      if filter[:range]
        filter_range_builder(id, filter)
        next
      end

      if filter[:required]
        unless params[filter[:param]]
          Rails::logger.error "Required filter #{filter[:param]} not given." unless params[filter[:param]]
          raise RequiredFilterNotProvided
        end
        filter[:value] = params[filter[:param]]
      else
        if filter[:required_combination]
          @null_filters[id] = nil unless params[filter[:param]]
          filter[:value] = params[filter[:param]]
          next
        end

        if (params[filter[:param]].is_a? Array) ?
                                          params[filter[:param]].all? {|p| p.blank? } :
                                          params[filter[:param]].blank?
          #dont remove default filters
          if filter[:value].nil?
            @filters_not_used.push(id)
          end
          next
        end
        filter[:value] = params[filter[:param]]
      end
    end
    @filters_not_used.each { |id| @filters.delete(id) }
  end

  def calc_next_offset
    @next_offset = @filters[:offset][:value].to_i + @filters[:limit][:value].to_i
    if @next_offset >= @query.limit(nil).offset(0).count
      @next_offset = -1
    end
  end

  def selected_column_name
    selected_columns.map do |column|
      column = column.is_a?(Hash) ? column[:return_name] : column
      next column unless column.include?('.')
      column.split('.')[-1]
    end
  end

  # Groups records by an attribute of the record
  # Can be used in conjunction with aggregate to aggregate a subset of records into an array
  def group
    return unless @group_by
    new_models = {}
    @models.each do |model|
      index = model.delete(@group_by)
      new_models[index] = new_models[index] || model
      if @aggregate_columns
        new_models[index][@aggregation_name] = new_models[index][@aggregation_name] || []
        hash = {}
        @aggregate_columns.each do |column|
          hash[column] = model.delete(column)
        end
        new_models[index][@aggregation_name] << hash
      end
    end
    @models = new_models.map {|k,model| model}
  end

  def columns
    # Columns can be objects, in the case where the name and the column are not the same
    return *selected_columns if selected_columns.all? {|s_c| s_c.is_a? String}
    selected_columns.map {|s_c| s_c.is_a?(String) ? s_c : s_c[:column]}
  end

  # We use pluck instead of select because select will keep id within the model, and we want complete control on what
  # columns are returns. This is also much faster than selecting and then doing a map to get only the selected columns.
  def pluck_to_models
    if @prior_model && @filters[:offset][:value] == 0
      @query = @model.from("(#{@query.to_sql} UNION #{@prior_model.to_sql}) AS #{@prior_model.table.name}").order(time: @filters[:order][:value])
    end

    #add some metrics for this part, since its the most intensive part (data loading, column formatting)
    data = []
    duration = collect_metrics do
      data = @query.pluck(*columns)
    end
    @metrics[:get_attributes] = "Loading attributes from the database took #{duration} seconds"

    duration = collect_metrics do
      @models = data.map{|pa| Hash[selected_column_name.zip(pa)]}
      apply_special_formatting
    end

    @metrics[:column_format] = "Formatting the returned data took #{duration} seconds"
  end

  def apply_special_formatting
    # Machine states duration are big decimal which rails converts to string automatically when passed to Rails JSON engine
    # Force it to float so that to_json does not transform it to a string.
    if @model.klass == MachineState
      @models.map do |state|
        unless state["calculated_duration"].nil?
          state["calculated_duration"] = state["calculated_duration"].to_f
        end
        state
      end
    end
  end

  # Pulled from RxModule ApplicationController since we're cutting some fat by only including The API class.
  def find_arel(base, query)
    path = query.split(".")
    if path.length == 1
      base.arel_table[query]
    else
      path[0..-2].each do |part|
        # queries on a joined table are plural, but the accessor could be singular
        possible_names = [part.singularize, part.pluralize]
        reflection = possible_names.map { |name| base&.reflect_on_association(name) }.compact.first
        base = reflection&.klass
      end
      find_arel(base, path[-1]) unless base.nil?
    end
  end

  def model_side_num_to_L_R
    @models.each do |model|
      next if model["side"].nil?
      model["side"] = Sideable.lens_side_num_to_L_R(model["side"])
    end
  end
end
