class JobStateFinishAbortRatioDashboard < RxModule::Dashboards::DualAxisChart
  include MachineDashboardHelpers
  include DataHelpers
  include KpiHelpers
  include BucketCalculator

  expose('job_state_finish_abort_ratio', 'Job State Finish/Abort Ratio', 'View the percentage of finished vs aborted job states for the selected machine(s)')
  machine_performance_setup(type_override: 'bar', options_override: {
    chart: {
      stacked: true,
      stackType: "100%"
    },
    tooltip: {
      enabledOnSeries: [0, 1],
      formatter: 'finish_abort_ratio'
    },
    legend: {
      position: 'top',
      horizontalAlign: 'right',
      customLegendItems: %w[Finished Aborted]
    },
    plotOptions: {
      bar: {
        horizontal: true
      }
    },
    dataLabels: {
      enabled: true,
      formatter: "finish_abort_ratio"
    },
    colors: %w[#3F47CC #ED1B24 #C3C3C3],
    xaxis: {
      title: {
        text: "Ratio"
      },
      labels: {
        formatter: "percent"
      }
    },
    yaxis: {
      title: {
        text: "Machine"
      }
    }
  })
  parameter(:split_machine_groups, "Split machine group into individual machines?", "When enabled, machine groups will be split into individual machines", :boolean, default: false)

  def unroll_target?(target)
    split_machine_groups && target.is_a?(MachineGroup)
  end

  def build_results!
    setup_targets!

    setup_series!

    s_time, e_time = start_and_end_time

    all_machine_ids = machine_id_groups.map { |mig| mig[:machine_ids] }.flatten.uniq
    job_states_per_machine = MachineKpi.job_state_breakdown_between(s_time, e_time, Timeframe::ALL, all_machine_ids, options: { aggregate_machines: false })

    machine_id_groups.each do |machine_id_group|
      finished_total = machine_id_group[:machine_ids].reduce(0) { |sum, k| sum + job_states_per_machine[k][:finished][:count].to_i }
      aborted_total = machine_id_group[:machine_ids].reduce(0) { |sum, k| sum + job_states_per_machine[k][:aborted][:count].to_i }
      push_point("finished", machine_id_group[:x_axis_name], finished_total)
      push_point("aborted", machine_id_group[:x_axis_name], aborted_total)
      push_point("no_data", machine_id_group[:x_axis_name], aborted_total + finished_total == 0 ? 1 : 0)
    end
  end

  def process_combos!
    # override no-op
  end

  def machine_id_groups
    @_machine_id_groups ||= []
  end

  def setup_targets!
    @_machine_id_groups = parsed_targets.map do |target|
      if target.is_a?(MachineGroup)
        if split_machine_groups
          target.machines.map do |sub_target|
            {
              x_axis_name: MachineDashboardHelpers.target_to_name(sub_target),
              machine_ids: [sub_target.id]
            }
          end
        else
          {
            x_axis_name: MachineDashboardHelpers.target_to_name(target),
            machine_ids: target.machine_ids
          }
        end
      else
        {
          x_axis_name: MachineDashboardHelpers.target_to_name(target),
          machine_ids: [target.id]
        }
      end
    end.flatten
  end

  def setup_series!
    metric = ProductionSetting.metric_is_pair? ? "jobs" : "lens"

    finished_series = push_series("finished", "Finished")
    finished_series[:type] = 'bar'
    finished_series[:metric] = metric

    aborted_series = push_series("aborted", "Aborted")
    aborted_series[:type] = 'bar'
    aborted_series[:metric] = metric

    no_data_series = push_series("no_data", "No Data")
    no_data_series[:type] = 'bar'
    no_data_series[:metric] = metric
  end

  def split_machine_groups
    @parameters[:split_machine_groups]
  end
end