class MachinePerformanceOverviewDashboard < RxModule::Dashboards::DualAxisChart
  include MachineDashboard
  include DataHelpers
  include MachineConcern
  include KpiHelpers
  include BucketCalculator

  ui_hint(type: 'type_override')
  ui_hint(type_override: 'machine_performance_overview')
  expose('machine_performance_overview', 'Machine Performance Overview', 'View machine performance in a graph.')
  parameter(:buckets, "Hours to display", "How many hours to display", :string, required: true, value_list: -> { total_time })
  parameter(:offset, "Start at mid-hour?", "When enabled, column starts with start at the middle of the hour (ex: 1:30, 2:30, etc.)", :boolean, default: false)
  targets("Machine", "The machine(s) to track", -> { machine_targets }, custom_targets: false)

  def build_results!
    setup_targets!

    addons = [:as_percent]

    s_time, e_time = start_and_end_time

    machine_ids = machines.pluck(:id)

    machine_states = MachineKpi.state_breakdown_between(s_time, e_time, Timeframe::HOUR_BY_HOUR, machine_ids, options: { addons: addons })

    waitingforjobs = extract_target_job_state(machine_states, :waitingforjobs)
    processing = extract_target_job_state(machine_states, :processing)

    throughput = MachineKpi.throughput_between(s_time, e_time, Timeframe::HOUR_BY_HOUR, machine_ids, options: {})

    build_dashboard_series('processing', processing, 'bar', :format_processing)
    build_dashboard_series('waitingforjobs', waitingforjobs, 'bar', :format_waitingforjobs)
    build_dashboard_series('throughput', throughput, 'line', :format_throughput)
  end

  def process_combos!
    # override no-op
  end

  def sort_series
    false
  end

  def buckets
    return MachinePerformanceOverviewDashboard.total_time[0][:value] if @parameters[:buckets].nil?
    @parameters[:buckets]
  end

  def offset
    @parameters[:offset]
  end

  def parameters
    @parameters
  end

  def machines
    @_machines ||= []
  end

  def setup_targets!
    ids = []
    targets.each do |id|
      category, ref = id.split("-")
      case category
      when 'm'
        ids << ref.to_i
      when 'g'
        g = MachineGroup.find(ref)
        g.machine_ids.each {|m_id| ids << m_id } unless g.nil?
      end
    end
    ids = ids.uniq # could be duplicates between machine and group
    @_machines = Machine.where(id: ids).find_each.to_a
  end

  def start_and_end_time
    now_time = Time.zone.now
    total_seconds = (buckets.to_i - 1).hours
    s_time = now_time.advance(seconds: -1 * total_seconds).beginning_of_hour
    e_time = now_time.end_of_hour
    if offset
      if now_time.strftime("%M").to_i < 30
        s_time = s_time.advance(minutes: -30)
        e_time = e_time.advance(minutes: -30)
      else
        s_time = s_time.advance(minutes: 30)
        e_time = e_time.advance(minutes: 30)
      end
    end
    [s_time, e_time]
  end

  def extract_target_job_state(data, target_state)
    new_data = {}
    data.each do |time_bucket, data|
      new_data[time_bucket] = data.slice(target_state)
    end
    new_data
  end

  def build_dashboard_series(name, data, type, formatter)
    new_series = push_series("id-#{name}", name)
    new_series[:type] = type
    if new_series[:id] == "id-throughput"
      new_series[:metric] = ProductionSetting.metric_is_pair? ? "jobs" : "lens"
    end
    data.each do |time, result|
      push_point(new_series[:id], "#{time.strftime('%H:%M')} - #{(time + 1.hour).strftime('%H:%M')}", send(formatter, result))
    end
  end

  def format_processing(bucket_value)
    bucket_value[:processing][:percent]
  end

  def format_waitingforjobs(bucket_value)
    bucket_value[:waitingforjobs][:percent]
  end

  def format_throughput(bucket_value)
    bucket_value[:count]
  end

  class << self
    def run_in_background?
      true
    end

    def target_to_name(target)
      if target.is_a? Machine
        "#{target.name} (#{target.sn})"
      elsif target.is_a? MachineGroup
        target.name
      end
    end

    def target_to_id(target)
      if target.is_a? Machine
        "m-#{target.id}"
      elsif target.is_a? MachineGroup
        "g-#{target.id}"
      end
    end

    def machine_targets
      MachineGroup.all.map do |g|
        { label: target_to_name(g), value: target_to_id(g), values_in_group: g.machines.map {|m| target_to_id(m)}, group: "Machine Group", description: "Report on performance for the entire #{g.name} group" }
      end + Machine.all.map do |m|
        { label: target_to_name(m), value: target_to_id(m), group: "Machine", description: "Report on performance for machine #{m.name}, sn #{m.sn}" }
      end
    end

    def total_time
      [
        {
          label: '8',
          value: '8',
          description: 'Display 8 hours of machine data'
        },
        {
          label: '12',
          value: '12',
          description: 'Display 12 hours of machine data'
        }
      ]
    end
  end
end