class MachineStateRatioDashboard < RxModule::Dashboards::DualAxisChart
  include MachineDashboardHelpers
  include DataHelpers
  include KpiHelpers
  include BucketCalculator

  expose('machine_state_ratio', 'Machine State Ratio', 'View the percentage of each machine state for the selected machine(s)')
  machine_performance_setup(type_override: 'bar', options_override: {
    chart: {
      stacked: true,
      stackType: "100%"
    },
    tooltip: {
      formatter: 'machine_state_ratio'
    },
    legend: {
      position: 'top',
      horizontalAlign: 'right'
    },
    plotOptions: {
      bar: {
        horizontal: true
      }
    },
    dataLabels: {
      enabled: true,
      formatter: 'machine_state_ratio'
    },
    xaxis: {
      title: {
        text: "Ratio"
      },
      labels: {
        formatter: "percent"
      }
    },
    yaxis: {
      title: {
        text: "Machine"
      }
    }
  })
  parameter(:split_machine_groups, "Split machine group into individual machines?", "When enabled, machine groups will be split into individual machines", :boolean, default: false)

  def unroll_target?(target)
    split_machine_groups && target.is_a?(MachineGroup)
  end

  def build_results!
    setup_targets!

    s_time, e_time = start_and_end_time

    all_machine_ids = machine_id_groups.map { |mig| mig[:machine_ids] }.flatten.uniq
    state_totals_per_machine = MachineKpi.state_breakdown_between(s_time, e_time, Timeframe::ALL, all_machine_ids, options: { aggregate_machines: false, addons: [:as_percent] })
    sorted_series = init_sorted_series

    machine_id_groups.each do |machine_id_group|
      state_totals = {}
      machine_id_group[:machine_ids].each do |machine_id|
        state_totals_per_machine[machine_id].each do |state, val|
          state_totals[state] = 0 if state_totals[state].nil?
          state_totals[state] += val[:count]
        end
      end
      state_totals.each do |state, val|
        x = machine_id_group[:x_axis_name]
        y = val
        next if y <= 0
        series_id = state.to_s
        legend_str = state.to_s
        if ProductionSetting.use_simple_machine_state_colors?
          state_group = self.class.find_state_group_for_state(state)
          series_id = self.class.group_state_id(state)
          legend_str = state_group.nil? ? "Unknown state group" : state_group.join(" - ")
        end
        sorted_series[series_id] << { legend_str: legend_str, state: state, x: x, y: y }
      end
    end
    sorted_series.to_a.each do |series_id, points|
      points.each do |point|
        if get_series(series_id).nil?
          # when creating a series, each machine must get an x-point in order to display it in the stacked chart
          push_series(series_id, point[:legend_str], self.class.color_for_state(point[:state]))
          machine_id_groups.each { |mig| push_point(series_id, mig[:x_axis_name], 0) }
        end
        increment_point(series_id, point[:x], point[:y])
      end
    end
  end

  def process_combos!
    # override no-op
  end

  def machine_id_groups
    @_machine_id_groups ||= []
  end

  def setup_targets!
    @_machine_id_groups = parsed_targets.map do |target|
      if target.is_a?(MachineGroup)
        if split_machine_groups
          target.machines.map do |sub_target|
            {
              x_axis_name: MachineDashboardHelpers.target_to_name(sub_target),
              machine_ids: [sub_target.id]
            }
          end
        else
          {
            x_axis_name: MachineDashboardHelpers.target_to_name(target),
            machine_ids: target.machine_ids
          }
        end
      else
        {
          x_axis_name: MachineDashboardHelpers.target_to_name(target),
          machine_ids: [target.id]
        }
      end
    end.flatten
  end

  def split_machine_groups
    @parameters[:split_machine_groups]
  end

  private

  def init_sorted_series
    if ProductionSetting.use_simple_machine_state_colors?
      {
        "processing" => [],
        "waitingforjobs" => [],
        "online" => [],
        "stoppage" => [],
        "offline" => []
      }
    else
      {
        "processing" => [],
        "waitingforjobs" => [],
        "maintenance" => [],
        "on" => [],
        "setup" => [],
        "ready" => [],
        "estop" => [],
        "halt" => [],
        "notready" => [],
        "error" => [],
        "off" => []
      }
    end
  end
end