import React, {useCallback, useState} from "react";
import {observer} from "mobx-react-lite";
import {Model} from "@rx-suite/data";
import {ModelProvider} from "@rx-suite/ui/build";
import {MachineDataPushTargetModel} from "../../stores/MachineDataPushTargetStore";
import {Dialog, DialogTitle, DialogContent, Button, DialogActions} from "@material-ui/core";
import {t} from "@lingui/macro";
import {ButtonField, SelectField} from "@rx-suite/ui";
import {useMachineOptions} from "../../hooks/select_options/useMachineOptions";
import {MachineMachineDataPushTargetStore} from "../../stores/MachineMachineDataPushTargetStore";
import {useSnackbar} from "notistack";

interface MachineDataPushTargetDialogProps {
    model: MachineDataPushTargetModel,
    open: boolean,
    handleClose: () => void,
}

export const ExternalForwardingAddMachineDialog: React.FunctionComponent<MachineDataPushTargetDialogProps> = observer(({model, open, handleClose,}) => {
    const {enqueueSnackbar} = useSnackbar();
    const [machineIds, setMachineIds] = useState<string[]>([]);
    const [isSubmitting, setIsSubmitting] = useState<boolean>()

    const handleSubmit = useCallback(async (selectedMachineIds: string[]) => {
        setIsSubmitting(true);

        try {
            const resp = await model.addMachines(selectedMachineIds);
            enqueueSnackbar(`${t`Machines added successfully`}`, {variant: "success"});
            MachineMachineDataPushTargetStore.refreshData();
            wrappedHandleClose();
        } catch (err) {
            console.error(err);
            enqueueSnackbar(`${t`Failed to add machines`}`, {variant: "error"});
        }

        setIsSubmitting(false);
    }, [model])

    const wrappedHandleClose = React.useCallback(() => {
        setMachineIds([])
        handleClose();
    }, [handleClose, setMachineIds])

    const handleMachineIdsChange = useCallback((name: string | undefined, value: string[]) => {
        setMachineIds(value);
    }, [])

    return <ModelProvider value={model}>
        <Dialog open={open} onClose={wrappedHandleClose} aria-labelledby="form-dialog-title" maxWidth={"md"} fullWidth={true}>
            <DialogTitle id="form-dialog-title">{t`Add Machines`}</DialogTitle>
            <DialogContent>
                <SelectField onChange={handleMachineIdsChange} value={machineIds} options={useMachineOptions()} multiple/>
                <DialogActions>
                    <ButtonField
                        text={t`Close`}
                        onClick={wrappedHandleClose}
                        color={'secondary'}
                        variant={'text'}
                    />
                    <ButtonField
                        text={t`Submit`}
                        onClick={() => handleSubmit(machineIds)}
                        loading={isSubmitting}
                        color={'secondary'}
                        variant={'text'}
                    />
                </DialogActions>
            </DialogContent>
        </Dialog>
    </ModelProvider>
})