import * as React from 'react';
import {useState} from "react";
import ReactTable, {FinalState, RowInfo} from "react-table";
import {t} from "@lingui/macro";
import {SystemSettingModel, SystemSettingStore} from "@rx-suite/stores";
import Dialog from "@material-ui/core/Dialog";
import IconButton from "@material-ui/core/IconButton";
import CloseIcon from "@material-ui/icons/Close";
import {
    DialogContent,
    DialogTitle,
    makeStyles,
    Typography,
} from "@material-ui/core";
import {ProcessParamsTable} from "./ProcessParamsTable";
import {TextExpand} from "../generic/TextExpand";
import {JobStateInfo, Side, sideDisplayText} from "../../pages/job_history/JobHistoryOperatorForm";
import {JobStateName} from "../../stores/JobStateStore";

type JobStateTableProps = {
    machineName: string;
    side: Side;
    jobStates: ReadonlyArray<JobStateInfo>;
    isBoxCoater: boolean;
};

const useStyles = makeStyles((theme) => ({
    viewProcessData: {
        fontSize: '14px',
        fontWeight: 400,
        marginLeft: '0.5rem',
        color: 'blue',
        '&:hover': {
            cursor: 'pointer',
            opacity: 0.5,
            textDecoration: 'underline'
        }
    },
    machineName: {
        fontWeight: 'bold'
    },
    dialog: {
        margin: 0,
        padding: theme.spacing(2)
    },
    closeButton: {
        position: 'absolute',
        right: theme.spacing(1),
        top: theme.spacing(1),
        color: theme.palette.grey[500],
    },
    dialogContent: {
        padding: theme.spacing(2)
    }
}));


export const JobStateTable: React.FunctionComponent<JobStateTableProps> = ({ machineName, side, jobStates, isBoxCoater }) => {

    // display all the job states of a machine inside a table

    const classes = useStyles();

    const [dialogContent, setDialogContent] = useState<JSX.Element|null>(null);

    const [open, setOpen] = useState<boolean>(false);

    // dialog close func
    const handleClose = () => {

        setDialogContent(null);

        setOpen(false);
    };

    // on click to display the process params in a table with a title
    const viewProcessOnClick = (d: JobStateInfo) => {

        if (!d) return; // do nothing if side is empty

        const cell = d.state;

        if (!cell) return; // do nothing if side is empty

        const cellState = cell.state;

        if (cellState !== 'finished') return; // do nothing if state is not finished

        // if the cell is a finished state
        // display the process params

        const cellProcessParams = d.process_params;

        const dialogContent = <>
            <DialogTitle disableTypography>
                <Typography variant="h5">
                    <span>{t`Process data for:`}</span>
                    {' '}
                    <span className={classes.machineName}>{machineName}</span>
                    {' - '}
                    <span>{sideDisplayText(side, isBoxCoater)}</span>
                </Typography>
                <IconButton className={classes.closeButton} onClick={handleClose}>
                    <CloseIcon />
                </IconButton>
            </DialogTitle>
            <DialogContent className={classes.dialogContent} dividers>
                <ProcessParamsTable
                    processParams={cellProcessParams}
                />
            </DialogContent>
        </>;

        setDialogContent(dialogContent);

        setOpen(true);
    };

    // this callback is used when the comment/scrap reason/time is truncated due to length
    // it will display the full length of the string inside a dialog
    const viewAllCallback = (title: string, content: string) => {

        const dialogContent = <>
            <DialogTitle disableTypography>
                <Typography variant="h5">
                    {title}
                </Typography>
                <IconButton className={classes.closeButton} onClick={handleClose}>
                    <CloseIcon />
                </IconButton>
            </DialogTitle>
            <DialogContent className={classes.dialogContent} dividers>
                <Typography variant="body1">
                    {content}
                </Typography>
            </DialogContent>
        </>;

        setDialogContent(dialogContent);

        setOpen(true);
    };

    // compare method to order the job states by time
    const compareByLatestTimeAsc = (a: JobStateInfo, b: JobStateInfo) => {

        // if the time does not exist, move it to the end of the list
        if (!a.state.time) return 1;
        if (!b.state.time) return -1;

        // if the time exists, compare it to the other time
        if (a.state.time > b.state.time) return 1;
        if (b.state.time > a.state.time) return -1;

        return 0; // times are equal
    };

    // sort the job states by time

    const sortedJobStates = jobStates.slice();

    sortedJobStates.sort(compareByLatestTimeAsc);

    const characterLimit = 100;

    const columns = [{
        id: 'state',
        Header: t`State`,
        accessor: (d: JobStateInfo) => {

            const cellState = d.state.state;

            if (cellState === 'finished') {

                return <>
                    <span>{cellState}</span>
                    <span
                        className={classes.viewProcessData}
                        onClick={() => { viewProcessOnClick(d); }}
                    >
                        {t`View process data`}
                    </span>
                </>;
            }

            return (cellState || 'N/A');
        },
        style: {
            overflowWrap: 'break-word',
            whiteSpace: 'normal'
        }
    },
    {
        id: 'comment',
        Header: t`Comment`,
        accessor: (d: JobStateInfo) => {

            const content = (d.state.comment || '');

            return <TextExpand
                content={content}
                limit={characterLimit}
                onShowMoreCallback={() => { viewAllCallback(t`Comment`, content); }}
            />;
        },
    },
    {
        id: 'scrapReason',
        Header: t`Scrap Reason`,
        accessor: (d: JobStateInfo) => {

            const content = (d.state.state === 'aborted'
                    ? (d.state.scrap_reason || '')
                    : ''
            );

            return <TextExpand
                content={content}
                limit={characterLimit}
                onShowMoreCallback={() => { viewAllCallback(t`Scrap Reason`, content); }}
            />;
        },
    },
    {
        id: 'time',
        Header: t`Time`,
        accessor: (d: JobStateInfo) => {

            const time = d.state.time;

            const systemSettingModel = SystemSettingStore.selectedModel as SystemSettingModel;

            const content = (time ? systemSettingModel.printDateTime(new Date(time)) : '');

            return <TextExpand
                content={content}
                limit={25} // don't let the time break to multi on full size display (1920px)
                onShowMoreCallback={() => { viewAllCallback(t`Time`, content); }}
            />;
        }
    }];

    return <>
        <ReactTable
            data={sortedJobStates}
            columns={columns}
            showPagination={false}
            // provide all of the data to the table,
            // without this, the table could show 30 rows,
            // but only fill 20 rows with data
            defaultPageSize={sortedJobStates.length}
            minRows={sortedJobStates.length <= 0 ? 3 : sortedJobStates.length} // if empty, pad the table with 3 rows
            sortable={false}
            getTrProps={(state: FinalState<JobStateInfo>, rowInfo: RowInfo | undefined) => {

                // style the rows based on state

                if (!rowInfo) return {}; // do nothing if the row is undefined

                const originalData = rowInfo.original;

                if (!originalData) return {}; // do nothing if side is empty

                const cell = originalData.state;

                if (!cell) return {}; // do nothing if side is empty

                const cellState: JobStateName = cell.state;

                // not started is an rx-routing system state, should not be used here
                // notStarted: 'rgb(240,240,240, .4)'

                const backgroundColor = {
                    started: 'rgba(255,225,35, .4)',
                    aborted: 'rgba(239,55,19, .4)',
                    finished: 'rgba(60,183,29, .4)',
                    postponed: 'rgba(240,240,240, .4)',
                    skipped: 'rgba(240,240,240, .4)',
                    restarted: 'rgba(255,225,35, .4)'
                }[cellState];

                return {
                    style: {
                        backgroundColor: backgroundColor,
                        fontWeight: (cellState === 'finished' ? 'bold' : 400)
                    }
                };
            }}
        />
        <Dialog
            className={classes.dialog}
            open={open}
            onClose={handleClose}
            maxWidth='md'
            fullWidth
        >
            {dialogContent}
        </Dialog>
    </>;
};
