import * as React from "react";
import {observer} from "mobx-react-lite";
import {
    Button,
    CircularProgress,
    Grid,
    makeStyles,
    Typography
} from "@material-ui/core";
import {t} from "@lingui/macro";
import {SelectField} from "@rx-suite/ui";
import {MachineTypeModel} from "../../stores/MachineTypeStore";
import {GetApp, Info} from "@material-ui/icons";
import {ProductionSettingStore} from "../../stores/ProductionSettingStore";
import {grey} from "@material-ui/core/colors";
import {useSnackbar} from "notistack";


type MachineManualsViewProps = {
    type: MachineTypeModel
}

const useStyles = makeStyles(theme => ({
    labels: {
        padding: '7px 0',
        textAlign: 'left'
    },
    downloadButtons: {
        margin: '6px'
    },
    noManual: {
        color: 'rgb(120,120,120)',
        textAlign: 'center'
    },
    noManualIcon: {
        margin: '0px 10px'
    },
    dialogIcon: {
        width: '5.75rem',
        height: '5.75rem',
        color: grey[600],
    },
    text: {
        textAlign: 'center'
    },
    subText: {
        marginTop: theme.spacing(1),
        color: grey[600]
    }
}))

export const MachineManualsView: React.FunctionComponent<MachineManualsViewProps> = observer(({type}) => {
    const classes = useStyles()
    const {enqueueSnackbar} = useSnackbar();

    const settings = ProductionSettingStore.selectedModel;
    if (!settings.machine_manual_path || !type || type.no_machine_manuals) {
        let message = null;
        if (!settings.machine_manual_path) {
            message = t`Machine manuals have not yet been registered to this MES-360 installation. Please contact support.`
        } else if (!type) {
            message = t`This machine has no machine type assigned. Please contact support.`
        } else if (type.no_machine_manuals) {
            message = t`There are no manuals available for this machine type.`
        }
        return <Typography variant={'h4'} className={classes.noManual}>
            <Info className={classes.noManualIcon}/>
            {message}
        </Typography>
    }

    const [operatingLanguage, setOperatingLanguage] = React.useState<string>("en")
    const [maintenanceLanguage, setMaintenanceLanguage] = React.useState<string>("en")
    const [downloadingMaintenance, setDownloadingMaintenance] = React.useState<boolean>(false)
    const [downloadingOperating, setDownloadingOperating] = React.useState<boolean>(false)

    const manualLanguage = React.useCallback((manualType: string) => {
        let value = null
        let language = null
        if (manualType === "operating") {
            value = operatingLanguage
            language = type.manual_languages?.operating_languages
        } else if (manualType === "maintenance") {
            value = maintenanceLanguage
            language = type.manual_languages?.maintenance_languages
        }

        return <SelectField
            options={language}
            onChange={(_, val) => {
                if (type.has_operating_manuals && manualType === "operating") {
                    setOperatingLanguage(val)
                } else if (type.has_maintenance_manuals && manualType === "maintenance") {
                    setMaintenanceLanguage(val)
                }
            }}
            value={value}
        />
    }, [type, operatingLanguage, maintenanceLanguage])

    const downloadButton = React.useCallback((manualType: string) => {
        let downloading = null
        if (manualType === "operating") {
            downloading = downloadingOperating
        } else if (manualType === "maintenance") {
            downloading = downloadingMaintenance
        }

        return <Button
            type={"download"}
            color={"secondary"}
            onClick={e => downloadManual(manualType)}
            variant={"contained"}
            startIcon={downloading ? <CircularProgress size={24}/> : <GetApp/>}
            disabled={downloading}
            className={classes.downloadButtons}>
                {t`Download`}
            </Button>
    }, [type, operatingLanguage, maintenanceLanguage, downloadingOperating, downloadingMaintenance])

    const downloadManual = React.useCallback( async (manualType: string) => {
        const formattedManualType = manualType.charAt(0).toUpperCase() + manualType.slice(1)
        const setDownloadingState = manualType === "operating" ? setDownloadingOperating : setDownloadingMaintenance
        try {
            const manualLanguage = manualType === "operating" ? operatingLanguage : maintenanceLanguage
            setDownloadingState(true)

            await type.getManual(manualType, manualLanguage);

            setDownloadingState(false)
            enqueueSnackbar(`${formattedManualType} manual downloaded successfully`, {variant: "success"});
        } catch (e) {
            console.error(e)
            setDownloadingState(false)
            enqueueSnackbar(t`Downloading ${formattedManualType} manual failed. Please contact your administrator`, {variant: "error"});
        }
    }, [type, maintenanceLanguage, operatingLanguage]);

    return (
        <div>
            <Grid container spacing={2}>
                <Grid item xs={2}>
                    {type.has_operating_manuals && <Typography variant={'h6'} className={classes.labels}>{t`Operating`}</Typography>}
                    {type.has_maintenance_manuals && <Typography variant={'h6'} className={classes.labels}>{t`Maintenance`}</Typography>}
                </Grid>
                <Grid item xs={2}>
                    {type.has_operating_manuals && manualLanguage("operating")}
                    {type.has_maintenance_manuals && manualLanguage("maintenance")}
                </Grid>
                <Grid item xs={1}>
                    {type.has_operating_manuals && downloadButton("operating")}
                    {type.has_maintenance_manuals && downloadButton("maintenance")}
                </Grid>
            </Grid>
        </div>
    )
})