import * as React from 'react';
import {observer} from "mobx-react-lite";
import {LineChart, ReportComponentMetaData, TemplateReportComponent} from "@rx-suite/ui";
import {MACHINE_PERFORMANCE_OVERVIEW_COLORS, resolveMachinePerformanceSeriesName} from "../../../hooks/reports/machine_performance_overview/machinePerformanceOverviewUtilities";
import {useMachinePerformanceOverviewLegend} from "../../../hooks/reports/machine_performance_overview/useMachinePerformanceOverviewLegend";
import {t} from "@lingui/macro";

type MachinePerformanceReportProps = {
    component: {
        meta: ReportComponentMetaData;
        data: any;
    };
};

export const MachinePerformanceReport: React.FunctionComponent<MachinePerformanceReportProps> = observer(({component}) => {
    const data = component.data;

    const UPTIME_SERIES_INDEX = 1;
    const UTILIZATION_SERIES_INDEX = 0;

    const uptimeValueForDataPoint = React.useCallback((dataPointIndex: number) => {
        // sum the uptime and the processing of the series to get the utilization
        return data.series[0]['data'][dataPointIndex].y + data.series[1]['data'][dataPointIndex].y;
    },[data.series]);

    const throughputSum = React.useMemo(() => {
        if (!data || !data.series || !data.series[2]) {
            return 0;
        }
        return data.series[2].data.map(d => Number(d.y)).reduce((a: number, b: number) => a + b, 0);
    }, [data]);

    const metric = React.useMemo(() => {
        if (!data || !data.options || !data.options.metric) {
            return 'Jobs';
        }
        return data.options.metric;
    }, [data]);

    const seriesPointValue = (value, {seriesIndex, dataPointIndex}) => {
        if(seriesIndex === UPTIME_SERIES_INDEX){
            value = uptimeValueForDataPoint(dataPointIndex);
        }

        if(seriesIndex == UTILIZATION_SERIES_INDEX || seriesIndex == UPTIME_SERIES_INDEX){
            return `${value.toFixed(2)}%`;
        }

        return `${value} ${metric}`;
    }

    return (
        <TemplateReportComponent meta={component.meta}>
            <LineChart series={data.series}
                       options={
                           {
                               title: {
                                   text: t`Performance Overview`,
                                   align: 'center'
                               },
                               subtitle: {
                                   text: t`Total` + ` ${metric}: ${throughputSum}`,
                                   align: 'left',
                                   offsetX: 64,
                                   style: {
                                       fontSize: '14px',
                                       fontWeight: 'bold',
                                       color: 'black'
                                   }
                               },
                               chart: {
                                   stacked: true,
                                   toolbar: {
                                       show: false
                                   },
                                   zoom: {
                                       enabled: false
                                   }
                               },
                               tooltip: {
                                   y: {
                                       formatter: seriesPointValue,
                                       title: {
                                           formatter: resolveMachinePerformanceSeriesName
                                       },
                                   },
                                   fillSeriesColor: false,
                                   marker: false
                               },
                               legend: {
                                   position: 'bottom',
                                   markers: {
                                       width: 0
                                   },
                                   formatter: useMachinePerformanceOverviewLegend
                               },
                               xaxis: {
                                   type: 'category'
                               },
                               dataLabels: {
                                   enabled: true,
                                   enabledOnSeries: [2]
                               },
                               stroke: {
                                   width: 2
                               },
                               yaxis: [
                                   {
                                       min: 0,
                                       max: 100,
                                       floating: false,
                                       tickAmount: 5,
                                       forceNiceScale: false,
                                       decimalsInFloat: 0,
                                       seriesName: 'waitingforjobs',
                                       title: {
                                           text: t`Utilization/Uptime (%)`
                                       },
                                   },
                                   {
                                       show: false,
                                       min: 0,
                                       max: 100,
                                       floating: false,
                                       tickAmount: 5,
                                       seriesName: 'processing',
                                       forceNiceScale: false,
                                   },
                                   {
                                       seriesName: 'throughput',
                                       opposite: true,
                                       tickAmount: 5,
                                       decimalsInFloat: 0,
                                       title: {
                                           text: t`Throughput` + `(${metric})`
                                       }
                                   }
                               ],
                               colors: [
                                   function({ value, dataPointIndex}) {
                                       const target = data.options.targets.utilization ? data.options.targets.utilization : 0;

                                       if (value >= target) {
                                           return MACHINE_PERFORMANCE_OVERVIEW_COLORS.utilization.over;
                                       } else {
                                           return MACHINE_PERFORMANCE_OVERVIEW_COLORS.utilization.under;
                                       }
                                   },
                                   function({ dataPointIndex }) {
                                       const value = uptimeValueForDataPoint(dataPointIndex)
                                       const target = data.options.targets.utilization ? data.options.targets.uptime : 0;

                                       if (value >= target) {
                                           return MACHINE_PERFORMANCE_OVERVIEW_COLORS.uptime.over;
                                       } else {
                                           return MACHINE_PERFORMANCE_OVERVIEW_COLORS.uptime.under;
                                       }
                                   },
                                   'rgb(26,93,173)'
                               ]
                           }
                       }
                       height={'100%'}
            />
        </TemplateReportComponent>
    )
})