import * as React from 'react';
import {observer} from "mobx-react-lite";
import {LinearProgress, Typography} from "@material-ui/core";
import {CheckboxField, ModelForm, ModelInput, SelectFileField} from "@rx-suite/ui";
import {BlockReuseSettingModel, BlockReuseSettingStore, importBlockReUseData,} from "../stores/BlockReuseSettingStore";
import {t, Trans} from "@lingui/macro";
import {ServiceController} from "../components/reuse_server/service_controller";
import {ImportExport} from "@material-ui/icons";
import {useSnackbar} from "notistack";
import {Alert} from "@material-ui/lab";
import {useEffect} from "react";
import {runInAction} from "mobx";

export const BlockReuseSettings = observer(() => {
    const settings = BlockReuseSettingStore.selectedModel as BlockReuseSettingModel;
    const {enqueueSnackbar} = useSnackbar();
    const [importing, setImporting] = React.useState(false)

    useEffect(() => {
        // This useEffect makes it so when you refresh the page, if the table is still locked
        // the model will go back to updating which will make it so the form cant be submitted
        if(!settings) {
            return
        }
        if(settings.block_table_locked) {
            runInAction(() => {
                settings._state = 'updating';
            });
        } else {
            runInAction(() => {
                settings._state = 'done';
            });
        }
    }, [settings])

    const triggerImport = React.useCallback(async(acceptedFiles) => {
        const file = (acceptedFiles || [])[0];

        if (file) {
            const reader = new FileReader();
            setImporting(true);

            reader.onabort = () => {
                console.error('file reading was aborted', reader.error)
                setImporting(false);
            }
            reader.onerror = () => {
                console.error('file reading has failed', reader.error);
                setImporting(false);
            }
            reader.onload = async() => {
                try {
                    await importBlockReUseData(reader.result as string);
                    enqueueSnackbar(<Trans>Finished Importing</Trans>, {variant: "success"})
                } catch(err) {
                    console.error(err)
                    enqueueSnackbar(<Trans>Error in Importing</Trans>, {variant: "error"})
                }
                setImporting(false);
            }
            reader.readAsText(file)
        }

    }, [])

    if(!settings) return <LinearProgress />

    return(
        <ModelForm model={settings} suppressMakeCopy suppressImportExport>
            <ServiceController name='ReUse Server' online={settings.server_online} enabled={settings.enable_server} startAction={settings.startReuseServer} stopAction={settings.stopReuseServer}/>
            <Typography variant={'h5'}><Trans>Settings</Trans></Typography>
            <div>
                <CheckboxField name={'enable_server'} label={t`Enable ReUse server`} color={"secondary"} onChange={settings.handleCheckboxChange} checked={settings.viewModel.enable_server}/>
                <div>
                    <Typography variant={'caption'}><Trans>Note: changes must be saved before the server can be started.</Trans></Typography>
                </div>
            </div>
            <ModelInput name={"server_port"}/>
            <ModelInput name={"block_wash_after"} />
            <ModelInput name={"block_discard_after"} />
            <div>
                {importing &&
                <div>
                    <LinearProgress />
                    <Typography variant={'caption'}><Trans>Importing data... this may take several minutes.</Trans></Typography>
                </div>}

                {!importing &&
                <SelectFileField label={t`Import Block Data`}
                                 buttonProps={{
                                     startIcon: <ImportExport/>,
                                     tooltip: t`Import block pieces from an external ReUse Server`
                                 }}
                                 accept={`.csv`}
                                 multiple={false}
                                 onDrop={triggerImport}
                                 disabled={importing}
                />}
            </div>
            {settings.block_table_locked &&
                <Alert severity='info'>
                    <Trans>Existing blocks are currently being updated. This may take several minutes...</Trans>
                </Alert>
            }
        </ModelForm>
    )
});
