import * as React from 'react';
import {observer} from "mobx-react-lite";
import {Conditional, FormProps, ModelForm, ModelInput, ModelSelect} from "@rx-suite/ui";
import {t, Trans} from '@lingui/macro';
import {MachineGroupStore} from "../../../stores/MachineGroupStore";
import {HintList} from "../../../components/generic/HintList";
import {Alert} from "@material-ui/lab";
import {ErrorFilterModel} from "../../../stores/ErrorFilterStore";
import {Grid, Typography} from "@material-ui/core";
import {ToggleFilterSwitch} from "../../../components/generic/ToggleFilterSwitch";

export const ErrorFilterForm: React.FunctionComponent<FormProps> = observer(({model}) => {
    const codeFormatHintText = [
        {title: t`Single Code`, text: '100'},
        {title: t`Code Range`, text: '100-1000'},
        {title: t`Single and Range`, text: '100, 150-200, 300'}
    ]

    const behaviourOptionsHintText = [
        {title: t`Ignore`, text: 'The targeted error codes will not appear in reports or dashboards.'},
        {title: t`Rename`, text: 'The targeted error codes will appear with the new name in reports or dashboards.'}
    ]

    const behaviourOptions = [
        {label: t`Ignore` as string, value: 'ignore'},
        {label: t`Rename` as string, value: 'replace'}
    ]

    const machineGroupOptions = React.useMemo(()=> {
        return MachineGroupStore.models.sort((a,b) => a.name.localeCompare(b.name)).map(g => {
            return {
                value: g.getId(),
                label: g.name
            }
        })
    }, [MachineGroupStore.models])

    // Certain fields can only be modified on a new model
    const readOnly = !model.isNew;

    return (
        <ModelForm model={model} suppressImportExport suppressMakeCopy suppressWrapper>
            <Conditional condition={readOnly}>
                <Grid container>
                    <Typography variant={"h6"}>
                        <Trans>Filter Enabled</Trans>
                    </Typography>
                    <ToggleFilterSwitch model={model as ErrorFilterModel}/>
                </Grid>
                <Alert severity='info'>
                    <Trans>Custom Alert Rules cannot be edited after creation.</Trans>
                </Alert>
            </Conditional>
            <ModelInput name={'name'}/>
            <ModelSelect
                name={'filter_target_id'}
                label={t`Target Machine Group`}
                disabled={readOnly}
                loading={MachineGroupStore.isLoading}
                options={machineGroupOptions}
            />
            <ModelInput
                name={'error_codes'}
                label={t`Targeted Error Codes`}
                disabled={readOnly}
                hint={
                    {
                        title: t`Available Error Code Formats`,
                        children: <HintList items={codeFormatHintText}/>
                    }
                }
            />
            <ModelSelect
                label={t`Action to apply`}
                name={'behaviour'}
                options={behaviourOptions}
                disabled={readOnly}
                hint={
                    {
                        title: t`Available Error Code Formats`,
                        children: <HintList items={behaviourOptionsHintText}/>
                    }
                }
            />
            <Conditional condition={(model as ErrorFilterModel).isReplaceFilter}>
                <ModelInput name={'replace_name'} disabled={readOnly}/>
            </Conditional>
            <ModelInput
                label={t`Replacement Error Code`}
                name={'group_code'}
                disabled={readOnly}
                hint={
                    {
                        title: t`Replacement Error Code`,
                        text: t`Replaces the target error codes with this new error code. For example, if error 400-500 are targeted and replaced with code g400, then all errors between 400 and 500 will appear as error code g400.`
                    }
                }
            />
        </ModelForm>
    );
});