import {attribute, DataStore, Model, model} from '@rx-suite/data';
import * as yup from 'yup';
import {t} from "@lingui/macro";
import {action, makeObservable, observable, runInAction} from "mobx";

class AbortFilterStoreDataStore<T extends Model> extends DataStore<T> {
    @observable showArchived = false;

    async toggleArchived(include_archived: boolean) {
        if (this.fetching) {
            return;
        }
        runInAction(() => {
            this.showArchived = include_archived;
        });
        const filter = this.fetchParams.filtered.find(filter => filter.id === 'archived');
        if(!filter){
            this.fetchParams.filtered.push({id: 'archived', value: this.showArchived.toString()});
        }else{
            filter.value = this.showArchived.toString();
        }
        await this.getTableData(true);
    }

    constructor(model, transportLayer, permissionTarget, routingInfo, eager) {
        super(model, transportLayer, permissionTarget, routingInfo, eager);
        makeObservable(this);
    }
}

@model({
    transportLayer: "/api/abort_filters",
    permissionTarget: "AbortFilter",
    routingInfo: {
        list: '/settings/data_filtering',
        create: '/settings/data_filtering/create',
        read: '/settings/data_filtering/:id/read',
        update: '/settings/data_filtering/:id/edit'
    },
    storeClass: AbortFilterStoreDataStore
})

export class AbortFilterModel extends Model {
    @attribute({
        label: t`Rule Name`,
        validation: yup.string().required(t`Rule name is required`)
    })
    name: string = ''

    @attribute({
        label: t`Regex Filters`,
        validation: yup.string().required(t`At least one regex filter must be applied`)
    })
    regex: string = '';

    @attribute( {
        label: t`Replacement Name`,
        validation: yup.string().required(t`Replacement Name is required when selected Behaviour is Rename`)
    })
    replace_name: string = ''

    @attribute({
        label: t`Group Code`,
        validation: yup.string().required(t`Group Code is required`)
    })
    group_code: string = ''

    @attribute({
        label: t`Archived?`,
        autoForm: false
    })
    archived: boolean = false;

    @action.bound
    async archive(confirmationMessage?: string) {
        let confirmation = true;
        if (confirmationMessage) {
            confirmation = window.confirm(confirmationMessage);
        }
        if (confirmation) {
            this.deletable = true;
            this._state = 'updating';
            const resp = await fetch(`${AbortFilterStore.transportLayer.base}/${this.id}/archive`, {
                method: "PUT",
                headers: {'Content-Type': 'application/json'}
            })
            const data = await resp.json();
            if (!resp.ok) {
                throw data.error;
            }
            this.store.deleteModel(this.getId().toString());
            this.store.totalRecords -= 1;
            if (this.store.fetchParams.limit !== undefined && this.store.tableData.length < this.store.fetchParams.limit) {
                this.store.getTableData(true);
            }
        }
    }

    @action.bound
    async restore(confirmationMessage?: string) {
        let confirmation = true;
        if (confirmationMessage) {
            confirmation = window.confirm(confirmationMessage);
        }
        if (confirmation) {
            this._state = 'updating';
            const resp = await fetch(`${AbortFilterStore.transportLayer.base}/${this.id}/restore`, {
                method: "PUT",
                headers: {'Content-Type': 'application/json'}
            });
            const data = await resp.json();
            if (!resp.ok) {
                throw data.error;
            }
            this.store.getTableData(true);
            this._state = 'done';
        }
    }

    constructor() {
        super();
        super.setup();
    }
}

export const AbortFilterStore = AbortFilterModel.store as unknown as DataStore<AbortFilterModel>