class ForwardMachineDataJob < ApplicationJob
  queue_as :machine_data_forwarding

  def perform(machine_id, forwarding_model, forwarding_path, forwarding_method, params, data)
    @machine = Machine.left_outer_joins(:machine_type).find_by(id: machine_id)
    @forwarding_model = forwarding_model
    @params = params

    return if @machine.nil?

    push_targets = @machine.push_targets_for_path(forwarding_path)

    push_targets.each do |target|
      endpoint = target[forwarding_path]
      method = target[forwarding_method] ? target[forwarding_method].downcase.to_sym : :put

      next if endpoint.blank?

      data = data.merge(@params) if target.include_request_params

      data = data.merge({name: @machine.name}) if target.include_machine_name

      url = "#{target.base_url}/#{endpoint}"
      Rails.logger.info "Forwarding to #{url}. Data => #{data}"
      response = Faraday.send(method, url) do |req|
        req.params.merge! resolve_params
        req.headers.merge! resolve_headers(target)
        req.body = data.to_json
      end
      parsed_response = begin
                          Oj.load(response.body, symbol_keys: true)
                        rescue # If json parsing fails, pass through the body
                          response.body
                        end
      raise parsed_response[:error] || parsed_response[:err] || parsed_response.to_s || "Unknown error - consult logs" if response.status == 500
      parsed_response
    rescue => err
      Rails.logger.error "Failed to forward to #{url}. Data => : #{data}. Error: #{err}"
    end
  end

  def resolve_params
    {
      "sn" => @params[:sn],
      "date" => @params[:date],
      "time" => @params[:time]
    }
  end

  def resolve_headers(target)
    base = {
      'Content-Type' => 'application/json'
    }
    base.merge! resolve_custom_headers(target)
  end

  def resolve_custom_headers(target)
    custom_headers = {}

    target.machine_data_push_target_headers do |header|
      customer_header_resolvers.each do |customer_header|
        if header.value.include? "$[#{customer_header[:key]}]$"
          header.value = header.value.gsub(customer_header[:key], self.send(customer_header[:resolver]))
        end
      end
      custom_headers[header.header] = header.value
    end

    custom_headers
  end

  def customer_header_resolvers
    [
      {
        key: "RXNET-INTERFACE-NAME",
        resolver: :resolve_rxnet_interface_name
      },
      {
        key: "MACHINE-TYPE",
        resolver: :resolve_machine_type
      },
      {
        key: "MACHINE-NAME",
        resolver: :resolve_machine_name
      }
    ]
  end

  def resolve_machine_type
    @machine.machine_type&.category || "other"
  end

  def resolve_machine_name
    @machine.name || ""
  end

  def resolve_rxnet_interface_name
    {
      process_data: 'process-parameters',
      error: 'machine-error',
      machine_state: 'machine-state',
      job_state: 'job-state',
      event: 'event',
      maintenance: 'maintenance',
      calibration: 'calibration'
    }[@forwarding_model]
  end
end
