class GenerateTemplateReportJob < ApplicationJob
  queue_as :report

  class InvalidReportTarget < StandardError; end

  def max_attempts
    # Reports are heavy, so don't use the default 25 retry
    3
  end

  def perform(report)
    report_id = report[:id]
    report_target = report[:target]
    report_params = JSON.parse(report[:params],{ symbolize_names: true })

    report_class = RxModule::Reporting.get_report(report_target)

    raise InvalidReportTarget if report_class.nil?

    update_report_status(report_id, 'running')

    generate_report_start = Time.now.utc

    report_results = report_class.new(report_params).run

    generate_report_end = Time.now.utc
    time_to_generate = (generate_report_end.to_f - generate_report_start.to_f).round(3)
    total_component_generate = report_results[:components].map{|c| c[:meta][:generate_component_details][:time_to_generate]}.inject(0, &:+)
    overhead_generate = (time_to_generate - total_component_generate).round(3)
    report_results[:meta][:generate_report_details] = {
      :start => generate_report_start,
      :end => generate_report_end,
      :time_to_generate => {
        :total => time_to_generate,
        :overhead => overhead_generate,
        :components => total_component_generate
      }
    }

    RxModule::Communicator.call_module_method('rx_reports', "#{report_url_for_template(report_id)}/attach_results", :post, {result: report_results})
  rescue StandardError => e
    Rails.logger.warn "Report generation ran into one or more issues: #{e}"
    update_report_status(report_id, 'failed', e)
  end

  def update_report_status(id, status, msg=nil)
    RxModule::Communicator.call_module_method('rx_reports', "#{report_url_for_template(id)}/update_status", :put, {status:status, message: msg})
  end

  def report_url_for_template(id)
    "#{reports_base_url}/#{id}"
  end

  def reports_base_url
    '/api/template_reports'
  end
end
