class UpdateOrderJob < ApplicationJob
  queue_as :order_data

  # NOTE: Serialized params are pass through here instead of leveraging the usual model serialization,
  def perform(order_id, data_params, serialized_params = nil)
    ActiveRecord::Base.connection_pool.with_connection do
      updated_order = update_order_and_jobs(order_id, data_params)

      return if updated_order.nil?
      updated_order.broadcast_update_order # internal forwarding

      return if serialized_params.nil?
      forward(serialized_params) # external forwarding
    end
  end

  def update_order_and_jobs(order_id, data_params)
    order = FindHelpers::OrderFinders.find_active_by_number(order_id)
    if order.nil?
      Rails.logger.warn "Failed to find order: #{order_id}. Data => : #{data_params}."
      return
    end
    jobs = order.jobs
    to_create = []
    to_update = {}
    current_time = Time.now

    %w[l r].each_with_index do |side_letter, side_num|
      # Job data is always either do 'L' or do 'R', never do 'B'
      sided_data_params = data_params.select do |data|
        data[:side] == side_num
      end

      next if sided_data_params.empty? # no job data to update for this side

      job = jobs.where(side: side_num).first
      next if job.nil? # can't update if no job exists for this side

      job_data = job.job_data

      sided_data_params.each do |sided_data_param|
        job_datum = job_data.find { |job_datum| job_datum[:name] == sided_data_param[:name] }
        if job_datum.nil?
          to_create << {name: sided_data_param[:name], side: sided_data_param[:side], value: sided_data_param[:value], job_id: job.id, created_at: current_time, updated_at: current_time}
        else
          to_update[job_datum.id] = {:value => sided_data_param[:value]}
        end
      end
    end

    JobDatum.transaction do
      JobDatum.insert_all!(to_create) unless to_create.empty?
      JobDatum.update(to_update.keys, to_update.values) unless to_update.empty?
    end

    order
  end

  def forward(serialized_params)
    push_targets = MachineDataPushTarget.where.not(forwarding_path_sym => [nil, ""])

    return if push_targets.size == 0

    ForwardOrderDataJob.perform_later(forwarding_path_sym, forwarding_method_sym, serialized_params)
  end

  def forwarding_path_sym
    :update_order_path
  end

  def forwarding_method_sym
    :update_order_method
  end
end
