module KpiTableHandler
  class JobStateCountKpiHandler < KpiTableHandler::Base::KpiTableHandler
    include KpiHelpers

    class JobStateNotSpecified < StandardError; end

    attr_accessor :job_states

    def initialize(start_time, end_time, timeframe, machine_ids, params: {}, options: {})
      raise JobStateNotSpecified unless params.key?(:job_states)

      self.job_states = params[:job_states]

      super(start_time, end_time, timeframe, machine_ids, options: options)
    end

    def table
      KpiJobState
    end

    def side_num
      [0, 1]
    end

    def state_sides
      job_states.map do |state|
        job_state_side(state)
      end.flatten
    end

    def job_state_side(job_state)
      side_num.map do |side|
        "#{job_state}_#{side}"
      end
    end

    def base_select
      selects = super
      selects.concat(state_sides.map do |state|
        table.arel_table[state.to_sym].sum.as("\"#{state.to_s}\"")
      end)
    end

    def zeroed_base_select
      selects = super
      selects.concat(state_sides.map do |state|
        "0 as \"#{state.to_s}\""
      end)
    end

    def niled_base_select
      selects = super
      selects.concat(state_sides.map do |state|
        "NULL as \"#{state.to_s}\""
      end)
    end

    def base_group
      group = []
      group << table.arel_table[:machine_id] unless aggregate_machines?
      group
    end

    def kpi_to_extract
      # combine state lens side into one col (ex: (finished_0 + finished_1) as finished)
      states = job_states.map do |state|
        "(#{job_state_side(state).join(" + ")}) as #{state}"
      end.join(",")
      states += ", machine_id" unless aggregate_machines?
      states
    end

    def timeframe_select
      selects = job_states.map {|job_state| table.arel_table[job_state.to_sym].sum.as(job_state)}
      selects << table.arel_table[:machine_id] unless aggregate_machines?
      selects
    end

    def timeframe_group
      base_group
    end

    def col_name
      cols = job_states.map {|job_state| job_state.to_sym}
      cols << :machine_id unless aggregate_machines?
      cols
    end

    # record is still an active record at this point, so extract the values we want
    def record_value(record)
      result = {}

      job_states.each do |state|
        result[state.to_sym] = record[state.to_sym]
      end

      result
    end

    def hash_targets_col
      state_sides.map{|m| m.to_sym}
    end

    def apply_hash_kpi_operation(hash)
      aggregate_machines? ?
        hash_as_total_col(hash) :
        hash_as_total_non_aggregated_machines(hash)
    end

    private

    # Sum 0/1 columns to get the final value
    def hash_as_total_col(hash)
      hash.each do |time_bucket, data|
        temp = {}
        job_states.each do |state|
          next temp[state.to_sym] = nil if use_planned_production_time? && data.values.all?(nil)

          temp[state.to_sym] = data["#{state}_0".to_sym] + data["#{state}_1".to_sym]
        end
        hash[time_bucket] = temp
      end

      hash
    end

    def hash_as_total_non_aggregated_machines(hash)
      if timeframe == Timeframe::ALL
        hash.each do |machine_id, buckets|
          temp = {}
          job_states.each do |state|
            next temp[state.to_sym] = nil if use_planned_production_time? && buckets.values.all?(nil)

            temp[state.to_sym] = buckets["#{state}_0".to_sym] + buckets["#{state}_1".to_sym]
          end
          hash[machine_id] = temp
        end
        hash
      else
        hash.each do |machine_id, buckets|
          buckets.each do |time_bucket, data|
            temp = {}
            job_states.each do |state|
              next temp[state.to_sym] = nil if use_planned_production_time? && data.values.all?(nil)

              temp[state.to_sym] = data["#{state}_0".to_sym] + data["#{state}_1".to_sym]
            end
            hash[machine_id][time_bucket] = temp
          end
        end
        hash
      end
    end
  end
end