module KpiTableHandler
  class MachineErrorCountKpiHandler < KpiTableHandler::Base::KpiTableHandler
    # This would get all the KPIMachineErrorCount
    include KpiHelpers
    include BucketCalculator

    class GroupCodeNotSpecified < StandardError; end

    def initialize(start_time, end_time, timeframe, machine_ids, params: {}, options: {})
      super(start_time, end_time, timeframe, machine_ids, options: options)
    end

    def table
      KpiMachineErrorCount
    end

    def base_select
      selects = super
      selects << table.arel_table[:count].sum.as("count")
      selects
    end

    def zeroed_base_select
      selects = super
      selects << "0 as count"
      selects
    end

    def niled_base_select
      selects = super
      selects << "NULL as count"
      selects
    end

    def non_kpi_columns
      selects = super
      selects << table.arel_table[:group_code]
      selects
    end

    def base_group
      group = [
        table.arel_table[:group_code]
      ]
      group << table.arel_table[:machine_id] unless aggregate_machines?
      group
    end

    def kpi_to_extract
      query = "group_code, (count) as count"

      #TODO should probably be generic?
      query += ", machine_id" unless aggregate_machines?

      query
    end

    def col_name
      cols = [
        :group_code,
        :count
      ]

      #TODO: should be in base KPI?
      cols << :machine_id unless aggregate_machines?
      cols
    end

    def timeframe_select
      base_select
    end

    def timeframe_group
      base_group
    end

    # This KPI is a bit different in the sense that we have to manipulate the col select to return a hash representing the group code and it's count
    def hash_col_select(bucket)
      { bucket[:group_code] => bucket[:count] }
    end

    def get_results_from_query_all(query)
      results = {}
      query.each do |record|
        if aggregate_machines?
          results.merge!({
                           record[:group_code] => record[:count]
                         }) { |k, a, b| a + b}
        else
          results[record[:machine_id]] ||= {}
          results[record[:machine_id]].merge!({
                                                record[:group_code] => record[:count]
                                              }) { |k, a, b| a + b}
        end
      end
      results
    end

    def all_cache_entries_exist?(from_sub_query, start_time, end_time)
      required_cache_entry_count = (((end_time - start_time) / KpiHelpers.cache_entry_size).ceil) * machine_ids.size

      unique_bucket_query(start_time, end_time).count == required_cache_entry_count
    end

    # This KPI has more than one bucket per time frame potentially, so we need a different exist? handler
    def unique_bucket_query(start_time, end_time)
      base = table.select(:time, :machine_id)
                  .group(kpi_table_class.arel_table[:time], :machine_id)
                  .where(kpi_table_class.arel_table[:time].gteq(start_time))
                  .where(kpi_table_class.arel_table[:time].lt(end_time))
                  .where(kpi_table_class.arel_table[:machine_id].in(machine_ids))

      table.select(:time).from(sub_table_to_from_string(base)).order(:time)
    end

    def find_missing_timeframes_query(from_sub_query, start_time, end_time)
      unique_bucket_query(start_time, end_time)
    end

    # This KPI also has a unique way of extracting the results of one cache entry
    def extract_result_from_cache_entry(record, results)
      time = record[:time]
      unless timeframe == Timeframe::HOUR_BY_HOUR
        time = time.advance(seconds: -1 * time.utc_offset) # reverse rails utc offset conversion (Already done db side)
      end
      group_code = record[:group_code]
      count = record[:count]
      if aggregate_machines?
        results[time] = {} if results[time].nil?
        results[time][group_code] = 0 if results[time][group_code].nil?
        results[time][group_code] += count
      else
        machine_id = record[:machine_id]
        results[machine_id] = {} if results[machine_id].nil?
        results[machine_id][time] = {} if results[machine_id][time].nil?
        results[machine_id][time][group_code] = 0 if results[machine_id][time][group_code].nil?
        results[machine_id][time][group_code] += count
      end
    end
  end
end