module MachineKpiCalculator

  # Base class for all Machine KPIs
  class MachineKpiCalculator

    class InvalidKpiHandler < StandardError; end

    include KpiHelpers
    include PlannedProductionTimeWithinShiftCalculator
    include UsePlannedProductionTime

    attr_accessor :start_time, :end_time, :timeframe, :machine_ids, :options, :data

    class << self
      def get_for_period_and_machines(start_time, end_time, timeframe, machine_ids, options: {})
        raise NotImplementedError
      end
    end

    def force_all(options)
      options[:force_all]
    end

    def initialize(start_time, end_time, timeframe, machine_ids, options)
      self.start_time = start_time
      self.end_time = end_time
      self.timeframe = timeframe
      self.machine_ids = machine_ids
      self.options = options
      self.data = {}
    end

    def aggregate_machines?
      !!options[:aggregate_machines]
    end

    def sum_as
      options[:sum_as]
    end

    def run
      build_data

      if aggregate_machines?
       @results = build_results
      else
        @results = build_results_per_machine
      end

      format_results

      @results
    end

    private

    # Overwrite this method to provide a custom output to a machine KPI
    # This must be done if the KPI does not return data in the format of timebucket => value
    def format_results
      wrap_with_count

      return if options[:addons].nil?

      options[:addons].map do |addon|
        method(addon).call(@results)
      end
    end

    def wrap_with_count
      if aggregate_machines?
        @results.each do |time, bucket|
          if bucket.is_a?(Hash)
            bucket.each do |key, value|
              @results[time][key] = {count: value}
            end
          else
            @results[time] = {count: bucket}
          end
        end
      else
        @results.each do |machine_id, hash|
          hash.each do |time, bucket|
            if bucket.is_a?(Hash)
              bucket.each do |key, value|
                @results[machine_id][time][key] = {count: value}
              end
            else
              @results[machine_id][time] = {count: bucket}
            end
          end
        end
      end
    end

    def as_percent(results)
      if aggregate_machines?
        percent_buckets(results, machine_count)
      else
        results.each do |machine_id, buckets|
          results[machine_id] = percent_buckets(buckets, 1)
        end
        results
      end
    end

    def planned_production_time(results)
      if aggregate_machines?
        planned_production_time_bucket(results)
      else
        results.each do |machine_id, buckets|
          results[machine_id] = planned_production_time_bucket(buckets)
        end
        results
      end
    end

    def percent_buckets(row, m_count)
      if timeframe == Timeframe::ALL
        bucket_size = bucket_size_for_timeframe(nil)
        row.each do |ck, cv|
          next row[ck][:percent] = 0 if bucket_size == 0
          row[ck][:percent] = ((cv[:count] / (bucket_size * m_count)) * 100).to_f.round(2)
        end
      else
        row.each do |time, col|
          bucket_size = bucket_size_for_timeframe(time)
          col.each do |ck, cv|
            next row[time][ck][:percent] = 0 if bucket_size == 0
            row[time][ck][:percent] = ((cv[:count] / (bucket_size * m_count)) * 100).to_f.round(2)
          end
        end
      end
    end

    def planned_production_time_bucket(row)
      if timeframe == Timeframe::ALL
        planned_production_time = bucket_size_for_timeframe(nil)
        row.each do |ck, cv|
          row[ck][:planned_production_time] = planned_production_time
        end
      else
        row.each do |time, col|
          planned_production_time = bucket_size_for_timeframe(time)
          col.each do |ck, cv|
            row[time][ck][:planned_production_time] = planned_production_time
          end
        end
      end
    end

    def machine_count
      machine_ids.size
    end

    # Handles calling the the handlers one at a time and dumping the result of each handler in data
    def build_data
      handlers.each do |handler|
        to_call = handler[:handler]
        key = handler[:key]

        if to_call.is_a?(Class) && to_call.method_defined?(:kpi_table_class)
          # to_call is a KPI handler, run it now
          data[key] = to_call.new(start_time, end_time, timeframe, machine_ids, options: options, params: handler[:params]).run
        elsif  to_call.is_a? Symbol
          data[key] = method(to_call).call
        else
          raise InvalidKpiHandler
        end
      end
    end

    # Array of KPI handlers or method to call to build the data required to build the result.
    # The key of the handler is how it will be saved in the data accumulator.
    # The params will be passed as-is to the KPI handler
    #
    # The handler can be a KpiHandler or the symbolized name of a method defined in the KPI itself
    # ex:
    # [
    #     {key: :downtime, handler: DowntimeKpiHandler, params: {:machine_states: ['estop', 'halt', 'error']}},
    #     {key: :number_failure, handler: :find_number_of_failures}
    # ]
    def handlers
      raise NotImplementedError
    end

    def handlers_per_machine
      raise NotImplementedError
    end

    # This method should use the data built by the handlers and return the final result of the KPI
    def build_results
      raise NotImplementedError
    end

    def build_results_per_machine
      raise NotImplementedError
    end

    def merge_with_addons!(key, val1, val2)
      val1.keys.map {|k| {k => (val1[k] || 0) + (val2[k] || 0)}}.reduce(:merge)
    end

    def time_to_next_30_minute_increment(time)
      increment = 30.minutes
      return Time.zone.at((time.to_f / increment).ceil * increment)
    end

    # If the requested timeframe is 'now' we have to truncate the timeframe to the nearest bucket and add the current bucket size
    def bucket_size_for_timeframe(time)
      return bucket_size_for_timeframe_ppt(time) if use_planned_production_time?

      # In the case where our bucket start time is earlier then the first state ever of the machine, we consider time spent before this state to be unknown,
      # so we use the bucket timeframe to calculate the percentage anyways. Example: time=00:00, processing at 12:00 with nothing before,
      # percentage of processing will still be 50%
      case timeframe
      when KpiHelpers::Timeframe::HOUR_BY_HOUR
        if time + 1.hour > Time.zone.now
          #TODO: test this case
          now_bucket_size        # Special case: hourly buckets may require an extra 30 min offset depending the current time
        else
          # AFAIK 1 and offset are the most granular time type, so cant have anything else than 1.hour bucket time
          return 1.hour
        end
      when KpiHelpers::Timeframe::DAY_BY_DAY
        # only activate if bucket day is same or greater than time now and end time is greater than time now bucket
        if time >= Time.zone.now.beginning_of_day && end_time >= KpiHelpers::kpi_bucket_time(Time.zone.now)
          return (KpiHelpers::kpi_bucket_time(Time.zone.now) - time.beginning_of_day) + KpiHelpers.machine_state_duration_current_bucket_size
        else
          # day by day will always span one day at a time unless its the now bucket, which is dealt above
          # We need to calculate actual time spent in bucket and not just bucket length
          if time < start_time
            # from midnight to bucket start is excluded
            1.day - start_time.seconds_since_midnight
          elsif time.day == end_time.day
            # Time spent in bucket
            end_time.seconds_since_midnight
          else
            1.day
          end
        end
      when KpiHelpers::Timeframe::WEEK_BY_WEEK
        # week by week will always span one week at a time unless its the now bucket
        if time.beginning_of_week(:sunday) == Time.zone.now.beginning_of_week(:sunday)
          return (KpiHelpers::kpi_bucket_time(Time.zone.now) - time.beginning_of_week(:sunday)) + KpiHelpers.machine_state_duration_current_bucket_size
        else
          return 1.week
        end
      when KpiHelpers::Timeframe::ALL
        if end_time > Time.zone.now
          return time_to_next_30_minute_increment(Time.zone.now) - start_time
        else
          return time_to_next_30_minute_increment(end_time) - start_time
        end
      end
    end

    def bucket_size_for_timeframe_ppt(time)
      case timeframe
      when KpiHelpers::Timeframe::HOUR_BY_HOUR
        if time + 1.hour > Time.zone.now
          return local_cache_get_active_time_between_for_shift_range(start_time, end_time, time, Time.zone.now)
        else
          return local_cache_get_active_time_between_for_shift_range(start_time, end_time, time, time + 1.hours)
        end
      when KpiHelpers::Timeframe::DAY_BY_DAY
        if time >= Time.zone.now.beginning_of_day && end_time >= KpiHelpers::kpi_bucket_time(Time.zone.now)
          return local_cache_get_active_time_between_for_shift_range(start_time, end_time, time.beginning_of_day, Time.zone.now)
        else
          # day by day will always span one day at a time unless its the now bucket, which is dealt above
          # We need to calculate actual time spent in bucket and not just bucket length
          if time < start_time
            # from midnight to bucket start is excluded
            return local_cache_get_active_time_between_for_shift_range(start_time, end_time, start_time, time + 1.day)
          elsif time.day == end_time.day
            # Time spent in bucket
            return local_cache_get_active_time_between_for_shift_range(start_time, end_time, time, end_time)
          else
            return local_cache_get_active_time_between_for_shift_range(start_time, end_time, time, time + 1.day)
          end
        end
      when KpiHelpers::Timeframe::WEEK_BY_WEEK
        # week by week will always span one week at a time unless its the now bucket
        if time.beginning_of_week(:sunday) == Time.zone.now.beginning_of_week(:sunday)
          return local_cache_get_active_time_between_for_shift_range(start_time, end_time, time.beginning_of_week(:sunday), Time.zone.now)
        else
          return local_cache_get_active_time_between_for_shift_range(start_time, end_time, time, time + 7.day)
        end
      when KpiHelpers::Timeframe::ALL
        if end_time > Time.zone.now
          return local_cache_get_active_time_between_for_shift_range(start_time, end_time, start_time, Time.zone.now)
        else
          return local_cache_get_active_time_between_for_shift_range(start_time, end_time, start_time, end_time)
        end
      end
    end

    def now_bucket_size
      offset = if now_bucket_requires_offset?
                 KpiHelpers.cache_entry_size
               else
                 0
               end

      KpiHelpers.machine_state_duration_current_bucket_size + offset
    end

    # If our start time is at the mid-hour (say 4h30)
    #   if it is currently 4h31 we don't need an offset (we are in the first half of the 'hour bucket').
    #   if it is currently 5h01 we need an offset (we are in the second half of the 'hour bucket').
    # If our start time is at the start of the hour (say 4h00)
    #    if it is currently 4h01 we don't need an offset (we are in the first half of the 'hour bucket').
    #    if it is currently 4h31 we need an offset (we are in the second half of the 'hour bucket').
    def now_bucket_requires_offset?
      return !KpiHelpers.time_is_in_second_half_hour?(Time.zone.now) if KpiHelpers.time_is_mid_hour(start_time)
      KpiHelpers.time_is_in_second_half_hour?(Time.zone.now)
    end
  end
end