module MachineKpiCalculator
  class OverallEquipmentEffectiveness < MachineKpiCalculator

    include KpiTableHandler
    include KpiHelpers

    class << self
      def get_for_period_and_machines(start_time, end_time, timeframe, machine_ids, options: {})
        kpi = OverallEquipmentEffectiveness.new(start_time, end_time, timeframe, machine_ids, options)
        kpi.run
      end
    end

    def initialize(start_time, end_time, timeframe, machine_ids, options)
      super(start_time, end_time, timeframe, machine_ids, options)
    end

    def handlers
      [
        {
          key: :uptime_percent,
          handler: KpiTableHandler::MachineStateDurationKpiHandler,
          params: {
            machine_states: %w[processing waitingforjobs],
            sum_as: 'uptime',
            as_percent: true
          }
        },
        {
          key: :job_states,
          handler: KpiTableHandler::JobStateCountKpiHandler,
          params: {
            job_states: %w[finished aborted],
            use_pair: false
          }
        },
        {
          key: :machine_state_duration,
          handler: KpiTableHandler::MachineStateDurationKpiHandler,
          params: {
            machine_states: %w[processing waitingforjobs]
          }
        },
        {
           key: :timeframe_overall_machine_states,
           handler: :timeframe_overall_machine_states,
           params: {}
        }
      ]
    end

    def timeframe_overall_machine_states
      if @timeframe_overall_machine_states.nil?
        @timeframe_overall_machine_states = MachineKpi.state_breakdown_between(start_time, end_time, Timeframe::ALL, machine_ids, options: {
          addons: [:as_percent],
          aggregate_machines: aggregate_machines?,
          use_planned_production_time: use_planned_production_time? ? :remove : false,
        })
      end

      @timeframe_overall_machine_states
    end

    def build_results
      # TODO: XBA (13/02/2023) - Not sure if this even useful for anywhere or something legacy. We seem to only use build result per machine. Not even sure this method is correct

      # Build Base KPI data
      results = calculate_oee_kpis

      # Extract KPI timelines to build OEE timeline
      base_kpi = machine_kpi_timeline(data)

      # Build OEE timeline (requires base KPI timeline first)
      results[:buckets] = {}

      machine_ids.each do |machine_id|
        results[:buckets][:machine_id] = build_oee_timeline(base_kpi[base_kpi])
      end

      results
    end

    def build_results_per_machine
      # Build Base KPI data
      results = calculate_oee_kpis

      # Group data per machine to simplify next steps
      grouped_base_data = group_data_per_machine

      # Build base KPI timeline per machine (Load, Yield, Uptime)
      grouped_base_kpi = {}

      machine_ids.each do |machine_id|
        grouped_base_kpi[machine_id] = machine_kpi_timeline(grouped_base_data[machine_id])
      end

      # Build OEE timeline (requires base KPI timeline first)
      results[:buckets] = {}

      machine_ids.each do |machine_id|
        results[:buckets][machine_id] = build_oee_timeline(grouped_base_kpi[machine_id])
      end

      results
    end

    def format_results
      as_percent
    end

    def as_percent
      if aggregate_machines?
        percent_buckets(@results[:buckets], machine_count)
      else
        @results[:buckets].each do |machine_id, buckets|
          percent_buckets(buckets, 1)
        end
      end
    end

    def percent_buckets(row, m_count)
      if timeframe == Timeframe::ALL
        bucket_size = bucket_size_for_timeframe(nil)
        return row = {percent: 0} if  bucket_size == 0
        row = {percent: ((row / (bucket_size * m_count)) * 100).to_f.round(2)}
      else
        row.each do |time, col|
          bucket_size = bucket_size_for_timeframe(time)
          next row[time] = {percent: 0} if  bucket_size == 0
          row[time] = {percent: ((col / (bucket_size * m_count)) * 100).to_f.round(2)}
        end
      end
    end

    private

    def calculate_oee_kpis
      uptime = calculate_timeframe_uptime
      lens_yield = calculate_timeframe_yield
      load = calculate_timeframe_load
      oee = uptime * lens_yield * load

      {
        uptime: format_kpi_value(uptime),
        yield: format_kpi_value(lens_yield),
        load: format_kpi_value(load),
        value: format_kpi_value(oee)
      }
    end

    def machine_timeframe_data_for_key(machine_id, key)
      # Each timeframe hash in index by the timeframe, so get the first value for the key
      data[key][machine_id]
    end

    def calculate_timeframe_uptime
      uptime = 0

      machine_ids.each do |machine_id|
        machine_states = machine_timeframe_data_for_key(machine_id,:timeframe_overall_machine_states)
        uptime += machine_states[:waitingforjobs][:percent] + machine_states[:processing][:percent]
      end

      adjust_for_machine_count(uptime) / 100 # as the % comes from the KPI handler, it is already multiplied by 100
    end

    def calculate_timeframe_yield
      overall_finished_lens = 0
      overall_aborted_lens = 0

      machine_ids.each do |machine_id|
        job_states =  data[:job_states][machine_id]

        overall_finished_lens += job_states.map{|k,v| v[:finished]}.sum
        overall_aborted_lens += job_states.map{|k,v| v[:aborted]}.sum
      end

      attempted_lens = overall_finished_lens + overall_aborted_lens

      overall_finished_lens / ensure_divider(attempted_lens.to_f) # Do not adjust for machine count - the divider does this naturally
    end

    def calculate_timeframe_load
      uptime = 0
      processing = 0

      machine_ids.each do |machine_id|
        machine_states =  machine_timeframe_data_for_key(machine_id,:timeframe_overall_machine_states)

        uptime += machine_states[:waitingforjobs][:percent] + machine_states[:processing][:percent]
        processing += machine_states[:processing][:percent]
      end

      processing / ensure_divider(uptime.to_f) # Do not adjust for machine count - the divider does this naturally
    end

    def group_data_per_machine
      data_by_machine = {}

      machine_ids.each do |machine_id|
        data_by_machine[machine_id] = {}
        data_by_machine[machine_id][:uptime_percent] = data[:uptime_percent][machine_id]
        data_by_machine[machine_id][:job_states] = data[:job_states][machine_id]
        data_by_machine[machine_id][:machine_state_duration] = data[:machine_state_duration][machine_id]
      end

      data_by_machine
    end

    def machine_kpi_timeline(machine_data)
      {
        uptime: uptime_from_buckets(machine_data[:uptime_percent]),
        yield: yield_from_buckets(machine_data[:job_states]),
        load: load_from_buckets(machine_data[:machine_state_duration])
      }
    end


    def uptime_from_buckets(buckets)
      res_buckets = {}
      buckets.each do |k,v|
        res_buckets[k] = v[:uptime] / 100
      end
      res_buckets
    end

    def yield_from_buckets(buckets)
      res_buckets = {}
      buckets.each do |k,v|
        aborted = (v[:aborted] || 0)
        finished = (v[:finished] || 0)
        attempted = aborted + finished
        res_buckets[k] = attempted == 0 ? 0 : finished / attempted.to_f
      end
      res_buckets
    end

    def load_from_buckets(buckets)
      res_buckets = {}
      buckets.each do |k,v|
        res_buckets[k] = calculate_load(v)
      end
      res_buckets
    end

    def calculate_load(bucket_value)
      processing = (bucket_value[:processing] || 0)
      waitingforjobs = (bucket_value[:waitingforjobs]  || 0)
      uptime = processing + waitingforjobs
      uptime == 0 ? 0 : processing / uptime.to_f
    end

    def build_oee_timeline(base_kpi_timeline)
      timeframes = base_kpi_timeline[:uptime].keys

      timeline = {}

      timeframes.each do |timeframe|
        timeline[timeframe] = (((base_kpi_timeline[:uptime][timeframe] || 0) * (base_kpi_timeline[:load][timeframe] || 0) * (base_kpi_timeline[:yield][timeframe] || 0)) * 100).round(2)
      end

      timeline
    end

    def format_kpi_value(value)
      (value * 100).round(2)
    end

    def adjust_for_machine_count(value)
      value / machine_ids.length.to_f
    end

    def ensure_divider(value)
      [value, 1].max
    end
  end
end