module MachineKpiCalculator
  class TopErrors < MachineKpiCalculator
    include KpiTableHandler
    include KpiHelpers
    include ErrorAndAbortPercentage

    class << self
      def get_for_period_and_machines(start_time, end_time, timeframe, machine_ids, options)
        kpi = TopErrors.new(start_time, end_time, timeframe, machine_ids, options)
        kpi.run
      end
    end

    def initialize(start_time, end_time, timeframe, machine_ids, options)
      super(start_time, end_time, timeframe, machine_ids, options)
    end

    def handlers
      [
        {
          key: :error_count,
          handler: KpiTableHandler::MachineErrorCountKpiHandler,
          params: {
          }
        }
      ]
    end

    def top_errors?
      !!options[:top_count]
    end

    def top_count
      options[:top_count]
    end

    def as_percent?
      options[:as_percent]
    end

    def extract_top_x(error_codes, top=5)
      top_errors = []
      error_codes.sort_by {|k, v| [-v, k]}.each do |err_code, value|
        break if top_errors.size >= top

        next if err_code == 'none'
        next if ignore_error_code?(err_code)

        top_errors << {code: err_code, count: value}
      end
      top_errors
    end

    def ignore_error_code?(err_code)
      return false if no_error_filter?(err_code)

      filter = ErrorFilter.find_by_group_code(err_code)
      return false if filter.nil?

      filter.is_ignore_filter?
    end

    def no_error_filter?(err_code)
      err_code.include?(KpiMachineErrorCount.no_filter_prefix)
    end

    def resolve_error_name(err_code)
      # First attempt to resolve a filter if any
      unless no_error_filter? err_code
        filter = ErrorFilter.find_by_group_code(err_code)

        unless filter.nil?
          error_code_prefix = if %w[- ,].any? { |sym| filter.error_codes.include?(sym)}
            "Multiple"
          else
            filter.error_codes
          end
          return "#{error_code_prefix} - #{filter.replace_name}" if filter.behaviour == 'replace'
        end
        err_no = err_code.split("_")[-1].to_i
      end

      # No filter, filter is invalid, or filter is 'use-as-is'
      err_no = err_code.split(KpiMachineErrorCount.no_filter_prefix)[-1].to_i if err_no.nil?

      # TODO: this will force all archived errors to become "Unknown Error"
      error_record = Error.find_by_errno(err_no)

      return 'Unknown Error' if error_record.nil? || error_record.error.nil?

      "#{err_no} - #{error_record.error}"
    end

    # group all error codes
    def group_codes(errors_hash)
      error_codes = {}
      errors_hash.each do |time, data|
        error_codes.merge!(data) { |k, a, b| a + b}
      end
      error_codes
    end

    def build_results
      # may need to filter out empty errors for ppt
      error_codes = group_codes(data[:error_count])
      extract_top_x(error_codes, 5)
    end

    # To get results per machine, we need to first sort all errors per their frequency
    # then we take the top 5 of those
    # then we can go per machine and select only those errors belonging to top 5.
    def build_results_per_machine
      total_error_count_hash = {}
      # Get all errors then sort them
      data[:error_count].each do |machine_id, buckets|
        if timeframe_all?
          total_error_count_hash.merge!(buckets) {|k, val1 , val2| val1 + val2}
        else
          buckets.each do |time, error_hash|
            total_error_count_hash.merge!(error_hash) {|k, val1 , val2| val1 + val2}
          end
        end
      end
      # only keep errors with total value greater than 0
      # values at zero can happen with ppt when none of the error happened within a shift but did happen outside of it.
      total_error_count_hash = total_error_count_hash.filter {|k,v| v > 0 }

      top_5_errors = extract_top_x(total_error_count_hash, 5).map {|error| error[:code]}

      results = {}
      # Extract per machine each error that belongs to top 5
      data[:error_count].each do |machine_id, buckets|
        results[machine_id] ||= {}
        if timeframe_all?
          get_errors_from_hash(buckets, results, top_5_errors, machine_id)
        else
          buckets.each do |time, error_hash|
            get_errors_from_hash(error_hash, results, top_5_errors, machine_id)
          end
        end
      end
      results
    end

    def get_errors_from_hash(error_hash, results, top_5_errors, machine_id, time = nil)
      top_5_errors.map do |error_code|
        error_name = resolve_error_name(error_code)
        if time
          results[machine_id][time][error_name] ||= 0
          # we want skipped errors to have 0 count
          next if error_hash[error_code].nil?
          results[machine_id][time][error_name] += error_hash[error_code]
        else
          results[machine_id][error_name] ||= 0
          # we want skipped errors to have 0 count
          next if error_hash[error_code].nil?
          results[machine_id][error_name] += error_hash[error_code]
        end
      end
    end
  end
end