# == Schema Information
#
# Table name: errors
#
#  id              :bigint           not null, primary key
#  errno           :integer
#  error           :string(4000)
#  errortype       :integer
#  machine_sn      :string(4000)
#  piececounter    :integer
#  status          :integer
#  time            :datetime
#  created_at      :datetime         not null
#  updated_at      :datetime         not null
#  error_filter_id :bigint
#  job_id          :bigint
#
# Indexes
#
#  index_errors_on_error_filter_id  (error_filter_id)
#  index_errors_on_job_id           (job_id)
#  index_errors_on_machine_sn       (machine_sn)
#
# Foreign Keys
#
#  fk_rails_...  (error_filter_id => error_filters.id)
#  fk_rails_...  (machine_sn => machines.sn) ON DELETE => cascade
#
class Error < ApplicationRecord
  include Pushable
  include ByMachineAndPeriod
  include ForwardMachineData

  enum errortype: [:unknown, :info, :alarm, :error]
  enum status: [:raised, :acknowledged, :finished]

  belongs_to :job, touch: true, required: false
  belongs_to :machine, foreign_key: :machine_sn, primary_key: :sn, touch: true

  belongs_to :error_filter, required: false

  def date_str
    time.strftime("%Y-%m-%d")
  end

  def time_str
    time.strftime("%T.%3N")
  end

  def forwarding_data
    {
      piececounter: piececounter,
      errorno: errno,
      error: error,
      errortype: errortype,
      status: status,
      job: job ? job.number : nil,
      lensside: job ? Sideable.side_letter_for_machine(job.side_num, machine) : nil,
    }
  end

  def forwarding_model
    :error
  end

  def forwarding_params
    {
      sn: machine_sn,
      date: time.strftime("%Y-%m-%d"),
      time: time.strftime("%H:%M:%S,%L")
    }
  end

  def has_filter?
    !error_filter.nil?
  end

  def self.delete_before(date)
    data_archive_settings = DataArchiveSetting.first
    max_delete_queue = data_archive_settings.delete_count_error
    return 0 if max_delete_queue.nil? || max_delete_queue <= 0
    deleted_objects = 0
    # Delete the errors that have existed before the given date
    # Only delete a maximum of `max_delete_queue` errors
    ActiveRecord::Base.connected_to(shard: :primary, role: :writing) do
      errors = Error.where(Error.arel_table[:time].lt(date)).limit(max_delete_queue)
      deleted_objects += errors.delete_all
    end
    deleted_objects
  end
end
