# == Schema Information
#
# Table name: job_states
#
#  id              :bigint           not null, primary key
#  abort_code      :integer
#  comment         :string(4000)
#  process         :string(4000)
#  scrap           :boolean
#  scrap_reason    :string(4000)
#  side            :integer
#  state           :integer          default("started")
#  substate        :string(4000)
#  time            :datetime
#  created_at      :datetime         not null
#  updated_at      :datetime         not null
#  abort_filter_id :bigint
#  batch_id        :string(4000)
#  job_id          :bigint           not null
#  machine_id      :bigint           not null
#
# Indexes
#
#  index_job_states_on_abort_filter_id  (abort_filter_id)
#  index_job_states_on_created_at       (created_at)
#  index_job_states_on_job_id           (job_id)
#  index_job_states_on_machine_id       (machine_id)
#  index_job_states_on_side             (side)
#  index_job_states_on_time             (time)
#
# Foreign Keys
#
#  fk_rails_...  (abort_filter_id => abort_filters.id)
#  fk_rails_...  (job_id => jobs.id) ON DELETE => cascade
#  fk_rails_...  (machine_id => machines.id) ON DELETE => cascade
#
class JobState < ApplicationRecord
  include Sideable
  include Pushable
  include ByMachineAndPeriod
  include ForwardMachineData
  include DataHelpers
  include BoxCoater::JobStateHelper

  enum state: JOB_STATES
  belongs_to :job, touch: true
  belongs_to :machine, touch: true
  belongs_to :abort_filter, required: false

  validates :state, presence: true

  def date_str
    time.strftime("%Y-%m-%d")
  end

  def datetime_str
    time.strftime("%Y-%m-%d %T.%3N")
  end

  def created_at_str
    created_at.strftime("%Y-%m-%d %T.%3N")
  end

  def forwarding_data
    {
      job: job.number,
      process: process,
      lensside: side_letter,
      state: state,
      substate: substate,
      comment: comment,
      scrap: scrap.nil? ? false : scrap,
      scrapReason: scrap_reason,
      batchId: batch_id,
      abortCode: abort_code
    }
  end

  def is_terminal?
    is_finished? || is_aborted?
  end

  def is_started?
    state == 'started'
  end

  def is_finished?
    state == 'finished'
  end

  def is_aborted?
    state == 'aborted'
  end

  def is_routing_kickout?
    is_aborted? && substate == 'AUTOFINISHED'
  end

  def has_filter?
    !abort_filter.nil?
  end

  def raw_abort_reason
    return comment unless comment.nil?
    return scrap_reason unless scrap_reason.nil?
    nil
  end

  def forwarding_model
    :job_state
  end

  def forwarding_params
    {
      sn: machine.sn,
      date: time.strftime("%Y-%m-%d"),
      time: time.strftime("%H:%M:%S,%L")
    }
  end

  # Returns the best estimate of the job state's process params (since there is no direct association between job_state and process_param)
  def process_params
    # Only job states that are finished have associated process params
    return [] unless state.upcase == "FINISHED"

    processes = JobProcess.where(machine_id: machine_id, job_id: job_id)

    best_guess = {
      process: nil,
      distance: Float::MAX
    }

    # Find the process (which is tied to process params) that is as close as possible to the job state. Time difference should be within 10 minutes.
    processes.each do |process|
      processTime = process.time.to_datetime.to_f
      jobStateTime = (time ? time.to_datetime.to_f : 0)
      timeDifference = (processTime - jobStateTime).abs

      next if timeDifference > 10.minutes

      if timeDifference < best_guess[:distance]
        best_guess = {
          process: process,
          distance: timeDifference
        }
      end
    end

    return [] if best_guess[:process].nil?

    best_guess[:process].job_process_parameters
  end
end
