# == Schema Information
#
# Table name: kpi_job_states
#
#  id          :bigint           not null, primary key
#  aborted_0   :integer          default(0)
#  aborted_1   :integer          default(0)
#  finished_0  :integer          default(0)
#  finished_1  :integer          default(0)
#  postponed_0 :integer          default(0)
#  postponed_1 :integer          default(0)
#  restarted_0 :integer          default(0)
#  restarted_1 :integer          default(0)
#  skipped_0   :integer          default(0)
#  skipped_1   :integer          default(0)
#  started_0   :integer          default(0)
#  started_1   :integer          default(0)
#  time        :datetime         not null
#  created_at  :datetime         not null
#  updated_at  :datetime         not null
#  machine_id  :bigint
#
# Indexes
#
#  index_kpi_job_states_on_machine_id        (machine_id)
#  kpi_kpi_job_states_uniq_time_per_machine  (time,machine_id) UNIQUE
#
# Foreign Keys
#
#  fk_rails_...  (machine_id => machines.id) ON DELETE => cascade
#
class KpiJobState < ApplicationRecord
  belongs_to :machine
  include KpiHelpers
  include DataHelpers

  def should_archive?
    false
  end

  def self.cache_job_state(job_state)
    cache_entry = KpiHelpers.get_kpi_cache_entry(KpiJobState, job_state.time, { machine_id: job_state.machine_id})
    column_name = "#{job_state.state}_#{job_state.side_num}"

    if cache_entry.nil?
      cache_entry = KpiJobState.create(time: KpiHelpers.kpi_bucket_time(job_state.time), machine_id: job_state.machine_id, column_name => 1)
    else
      cache_entry.update(column_name => cache_entry[column_name] += 1)
    end

    cache_entry
  end

  def self.state_side_key(state, side)
    ("#{state}_#{side}").to_sym
  end

  def self.backwards_cache_entry_calculation(entry, leave_empty: false, default_value: 0)
    start_time = entry[:time]
    end_time = start_time + 30.minutes
    machine_id = entry[:machine_id]

    cache_entry = empty_cache_entry(default_value)

    return wrap_cache_entry(cache_entry, start_time, machine_id) if leave_empty

    job_states = JobState.where(machine_id: machine_id)
                         .where(JobState.arel_table[:time].gteq(start_time).and(JobState.arel_table[:time]).lt(end_time))

    job_states.each do |job_state|
      cache_entry[state_side_key(job_state.state, job_state.job.side_num)] += 1
    end

    wrap_cache_entry(cache_entry, start_time, machine_id)
  end

  def self.wrap_cache_entry(entry, start_time, machine_id)
    [
      entry.merge!(
        {
          time: start_time,
          machine_id: machine_id.to_i,
          created_at: Time.zone.now,
          updated_at: Time.zone.now
        })
    ]
  end

  def self.empty_cache_entry(default_value = 0)
    cache_entry = {}
    %w[0 1].each do |side_num|
      JOB_STATES.each_key { |state| cache_entry[state_side_key(state.to_s, side_num)] = default_value}
    end
    cache_entry
  end
end

