# == Schema Information
#
# Table name: machines
#
#  id                  :bigint           not null, primary key
#  archived            :boolean          default(FALSE)
#  description         :string(4000)
#  is_conveyor_machine :boolean          default(FALSE)
#  name                :string(4000)
#  push_target_enabled :boolean          default(FALSE)
#  sn                  :string(4000)     not null
#  track_eer_metrics   :boolean          default(FALSE), not null
#  created_at          :datetime         not null
#  updated_at          :datetime         not null
#  machine_type_id     :bigint
#
# Indexes
#
#  index_machines_on_machine_type_id  (machine_type_id)
#  index_machines_on_sn               (sn) UNIQUE
#
# Foreign Keys
#
#  fk_rails_...  (machine_type_id => machine_types.id)
#
class Machine < ApplicationRecord
  scope :not_archived, -> { where(archived: false) }
  scope :blockers, -> { joins(:machine_type).where({machine_type: {category: 'blocker'}}) }
  include Pushable
  include MachineKpi
  include MachineStateDurationHelpers
  include MachineErrorHelpers
  include BoxCoater::MachineHelper
  include ProductionView
  include RxModule::LatestAssociatedModel

  has_many :machine_states, -> { order(:time) }, foreign_key: "machine_sn", primary_key: :sn
  has_many :problems, foreign_key: "machine_sn", primary_key: :sn, class_name: "Error"
  has_many :events, foreign_key: "machine_sn", primary_key: :sn, class_name: "Event"
  has_many :job_processes
  has_many :jobs, through: :job_processes

  has_many :job_states
  has_many :eer_yields
  has_many :kpi_job_states
  has_many :kpi_routing_kickouts

  has_and_belongs_to_many :machine_groups
  has_and_belongs_to_many :process_booking_workstations

  has_many :machine_machine_rejects
  has_many :machine_rejects, class_name: "Machine", through: :machine_machine_rejects

  has_many :machine_reject_machines, class_name: "MachineMachineReject", foreign_key: :machine_reject_id
  has_many :machines, through: :machine_reject_machines

  has_many :box_coater_events
  has_many :box_coater_batches

  belongs_to :machine_type, required: false

  has_many :machine_machine_data_push_targets
  has_many :machine_data_push_targets, :through => :machine_machine_data_push_targets

  def extended_name
    return sn if name.nil?
    return "#{name} (#{sn})"
  end

  def ensured_name
    name || sn || id
  end

  def current_state
    KpiCurrentMachineState.find_or_create_by(machine_sn: sn)
  end

  def has_machine_type?
    !machine_type_id.nil?
  end

  def belongs_to_group?
    machine_group_ids.size > 0
  end

  def translator_class
    return nil if self.machine_type.nil?
    self.machine_type.translator_class
  end

  #todo: Move to machine state duration KPI helper?
  def set_current_state(machine_state)
    cur_state = current_state
    # If the current machine state is in the future, the machine time sync was off, so replace with newer entry
    if (machine_state.time >= cur_state.time) || (current_state.time >= (Time.zone.now + 15.minutes))
      cur_state.update!(state: machine_state.state, time: machine_state.time, comment: machine_state.comment)
    end
  end

  def self.status_all
    status = {}
    last_machine_state_per_machine = Machine.latest_records_per_foreign_key(MachineState,
                                                            foreign_key_column: :machine_sn,
                                                            base_class_key_column: :sn,
                                                            time_column: :time,
                                                            limit: 1,
                                                            extra_select: [Machine.arel_table[:id].as('machine_id'),
                                                                           Machine.arel_table[:sn],
                                                                           Machine.arel_table[:name],
                                                                           Machine.arel_table[:is_conveyor_machine],
                                                                           Machine.arel_table[:machine_group_ids]

                                                            ],
                                                            extra_where: Machine.arel_table[:archived].eq(0),
                                                            from: add_machine_group_ids_to_machine_query(
                                                              extra_select: [Machine.arel_table[:is_conveyor_machine]],
                                                              extra_group: [Machine.arel_table[:is_conveyor_machine]]
                                                            ))

    last_machine_state_per_machine.each do |machine_machine_state|
      status[machine_machine_state.sn] = {
        machine_id: machine_machine_state.machine_id,
        name: machine_machine_state.name,
        groups: machine_machine_state[:machine_group_ids].nil? ? [] : machine_machine_state[:machine_group_ids].split(","),
        is_conveyor_machine: machine_machine_state.is_conveyor_machine
      }

      if machine_machine_state.state.nil?
        status[machine_machine_state.sn].merge!({
                                                  status: 'notconnected',
                                                  comment: nil,
                                                  time: Time.zone.now,
                                                  duration: 0,
                                                })
      else
        status[machine_machine_state.sn].merge!({
          status: DataHelpers::MACHINE_STATES[machine_machine_state.state],
          comment: machine_machine_state.comment,
          time: machine_machine_state.time,
          duration: Time.now - machine_machine_state.time,
        })
      end
    end
    status
  end

  def self.last_job_states(limit = 5)
    limit = ensure_limit(limit)
    job_states_per_machine = {}
    job_states = Machine.latest_records_per_foreign_key(JobState,
                                                        limit: limit,
                                                        extra_select: [
                                                          Machine.arel_table[:sn],
                                                          Machine.arel_table[:name],
                                                          Machine.arel_table[:machine_group_ids]
                                                        ],
                                                        extra_where: Machine.arel_table[:archived].eq(0),
                                                        from: add_machine_group_ids_to_machine_query)
    # using each here instead of find_each because the query is too complex for find_each and we are limited to 5 job_states per machine anyways.
    job_states.each do |job_state|
      job_states_per_machine[job_state.sn] ||= []
      attributes = job_state.attributes
      attributes["machine_group_ids"] = attributes["machine_group_ids"].nil? ? [] : attributes["machine_group_ids"].split(',')
      JobState.defined_enums.each do |enum_key, enum_value|
        attributes[enum_key] = enum_value.key(attributes[enum_key])
      end
      job_states_per_machine[job_state.sn] << attributes
    end
    job_states_per_machine
  end

  def self.ensure_limit(limit)
    limit = limit.nil? ? 5 : limit
    limit = limit.to_i > 1000 ? 1000 : limit.to_i
    limit
  end

  def self.machine_list
    options = []
    other = {label: 'other', options: []}
    Machine.find_each do |machine|
      if machine.has_machine_type?
        type = options.find {|c| c[:label] == machine.machine_type.category}
        if type.nil?
          type = {label: machine.machine_type.category, options: []}
          options << type
        end
        type[:options] << { label: machine.extended_name, value: machine.id }
      else
        other[:options] << { label: machine.extended_name, value: machine.id }
      end
    end
    options << other
    options
  end

  def self.add_machine_group_ids_to_machine_query(extra_select: [], extra_group: [])
    # This query's purpose is to add the column machine_group_ids to the machine records
    cast_type = if DatabaseServer.mssql?
                  'nvarchar'
                else
                  'varchar'
                end
    # Adding machine group ids as a column to the resulting query
    machine_groups_machines_table = Arel::Table.new(:machine_groups_machines)
    Machine.select(Machine.arel_table[:id],
                   Machine.arel_table[:name],
                   Machine.arel_table[:sn],
                   Machine.arel_table[:archived],
                   Arel::Nodes::NamedFunction.new('STRING_AGG',
                                                  [MachineGroup.arel_table[:id].cast(cast_type),
                                                   Arel::Nodes.build_quoted(',')])
                                             .as('machine_group_ids'),
                   *extra_select)
           .joins(Machine.arel_table
                         .join(machine_groups_machines_table, Arel::Nodes::OuterJoin)
                         .on(Machine.arel_table[:id].eq(machine_groups_machines_table[:machine_id]))
                         .join_sources)
           .joins(Machine.arel_table
                         .join(MachineGroup.arel_table, Arel::Nodes::OuterJoin)
                         .on(machine_groups_machines_table[:machine_group_id].eq(MachineGroup.arel_table[:id]))
                         .join_sources)
           .group(Machine.arel_table[:id],
                  Machine.arel_table[:name],
                  Machine.arel_table[:sn],
                  Machine.arel_table[:archived],
                  *extra_group)
  end

  concerning :LensKartReport do
    def self.each_machine_and_process_time_for_jobs(params)
      column_arr = params[:columns]

      state_columns = params[:columns].filter {|column_name| !(['order_number', 'order_time'].include? column_name)}
      fk_values = params[:orders]

      all_order_numbers = params[:orders].pluck('order_number').reject {|o| o === nil}

      # STEP 1 make a query that gets an order_number with its processors and the process_times.

      from_query =
        Order.select(
          Order.arel_table[:id],
          Order.arel_table[:job_number].as('order_number'),
          Order.arel_table[:created_at].as('order_time'),
          Job.arel_table[:id].as('job_id'),
          JobState.arel_table[:id].as('job_state_id'),
          JobState.arel_table[:time].as('finished_time'),
          JobState.arel_table[:machine_id].as('job_state_machine_id'),
          Machine.arel_table[:name].as('machine_name'),
          MachineType.arel_table[:category].as('machine_category'))
             .joins(:jobs => {:job_states => {:machine => :machine_type}})
             .where(JobState.arel_table[:state].eq(5))
             .where(Order.arel_table[:job_number].in(all_order_numbers))

      ternary_sql = lambda do |condition, result_true, result_false|
        Arel::Nodes::NamedFunction.new('IIF', [condition, result_true, result_false])
      end

      conditionally_aggregated_column = lambda do |column_name|
        category_name = column_name.split('_')[0].singularize
        col_value = column_name.include?('time') ? :finished_time : :machine_name

        ternary_sql.call(
          Order.arel_table[:machine_category].eq(category_name),
          Order.arel_table[col_value],
          Arel.sql('null')
        ).maximum.as(column_name)
      end

      window = Arel::Nodes::Window.new.partition([Order.arel_table[:id], Order.arel_table[:machine_category]]).order(Order.arel_table[:finished_time].desc)
      sub_query = Order.select(Arel::Nodes::NamedFunction.new('ROW_NUMBER', []).over(window).as('rn'), Arel.star).from("(#{from_query.to_sql}) as orders")

      step_1_sql = Order.select(
        state_columns.map do |column|
          conditionally_aggregated_column.call(column)
        end + [Order.arel_table[:order_number], Order.arel_table[:order_time]])
                        .from("(#{sub_query.to_sql}) as orders")
                        .where(Order.arel_table[:rn].lteq(1))
                        .group(Order.arel_table[:order_number], Order.arel_table[:order_time])

      #STEP 2 we need to put the above rows into a temp table

      temp_table_sql = [create_temp_table_sql]
      temp_table_sql += state_columns.map do |accessor|
        type = accessor.include?('time') ? datetime_type : 'varchar(40)'
        "#{accessor} #{type}"
      end + [' order_number varchar(40), order_time datetime']
      temp_table_sql = temp_table_sql.join(', ')
      step_2_sql = "#{temp_table_sql}) INSERT INTO #{temp_table_name} (#{(state_columns + ['order_number, order_time']).join(', ')})"

      sql_loaded_temp_table = "#{step_2_sql} #{step_1_sql.to_sql}"

      # STEP 3 Make a union for each order number request with its datetimereceived to properly select each unique orders.

      fk_processed = fk_values.map do |fk_hash|
        <<~SQL
SELECT #{column_arr.join(', ')}
FROM (
  SELECT *,
  ROW_NUMBER() OVER (PARTITION BY order_number ORDER BY order_time ASC) AS rn
  FROM #{temp_table_name}
  WHERE order_number = '#{fk_hash['order_number']}'
  AND order_time >= '#{fk_hash['datetimereceived']}'
) as orders
WHERE orders.rn <=1
      SQL
      end.join(' UNION ALL ')

      step_3_sql = "#{sql_loaded_temp_table}  #{fk_processed}"

      records = ActiveRecord::Base.connection.exec_query(step_3_sql)

      {
        hits: records,
      }
    end

    private

    def self.datetime_type
      if DatabaseServer.mssql?
        "datetime"
      else
        "timestamp"
      end
    end

    def self.create_temp_table_sql
      if DatabaseServer.mssql?
        "DECLARE #{temp_table_name} TABLE (idx INT IDENTITY(1,1)"
      else
        "CREATE TEMP TABLE #{temp_table_name} (idx INT"
      end
    end

    def self.temp_table_name
      if DatabaseServer.mssql?
        '@TempTable'
      else
        'TempTable'
      end
    end
  end

  def push_targets_for_path(path)
    MachineDataPushTarget.where(send_to: :all_machines).where.not(path => [nil, ""]) + self.machine_data_push_targets.where.not(path => [nil, ""])
  end

  # TODO: this should be retired once 'current_state' is fully implemented
  def last_state
    @last_state ||= machine_states.last
  end

  #TODO: delete?
  reportable("machine", "Machine", "Report on machines", {
    name: {
      name: "Name",
      desc: "The name of the machine"
    },
    sn: {
      name: "Serial Number",
      shortname: "SN",
      desc: "The serial number of the machine"
    },
    description: {
      name: "Description",
      desc: "The description of the machine"
    },
    is_conveyor_machine: {
      name: "Conveyor?",
      desc: "Whether or not the machine is a conveyor"
    },
  })
end
