module ReportMachines
  extend ActiveSupport::Concern

  class InvalidMachineTarget < StandardError; end

  def resolve_machine_ids
    raise InvalidMachineTarget if @parameters[:target_ids].blank?
    #Resolve machine target ids to machine or machine group, then into machine

    if @parameters[:target_ids].is_a? Array
      targets = @parameters[:target_ids].map {|target_id| target_from_id(target_id)}
      targets.map do |target|
        next target.id if target.is_a? Machine
        #If target is a machine group, get ids of all machines within
        target.machine_ids
      end.flatten.uniq
    else
      target = target_from_id(@parameters[:target_ids])
      return [target.id] if target.is_a? Machine
      #If target is a machine group, get ids of all machines within
      target.machine_ids
    end
  end

  def resolve_machine_targets(target_ids = @parameters[:target_ids])
    if target_ids.is_a? Array
      targets_from_id = target_ids.map {|target_id| target_from_id(target_id)}
      targets_from_id.map do |target|
        resolved_target_name(target)
      end
    else
      resolved_target_name(target_from_id(target_ids))
    end
  end

  def resolved_target_name(target)
    if target.is_a? Machine
      target.extended_name
    else
      target.name
    end
  end

  def target_from_id(id)
    type, ref = id.split('-')
    if type == 'm'
      Machine.find { |m| m.id.to_s == ref }
    elsif type == 'g'
      MachineGroup.find { |g| g.id.to_s == ref }
    end
  end

  class_methods do
    def use_machine_targets(multiple, include_groups = true)
      parameter(
        :target_ids,
        "Machine",
        "The machine that will be the target of the report",
        :string,
        multiple: multiple,
        required: true,
        value_list: -> {reporting_targets(include_groups)})
    end

    def reporting_targets(include_groups=true)
      options = []
      if include_groups
        group_option = {label: 'Groups', options: []}
        MachineGroup.find_each do |machine_group|
          group_option[:options] << { label: target_to_name(machine_group), value: target_to_id(machine_group), chart_type: target_to_chart_type(machine_group) }
        end
        options << group_option
      end
      other = {label: 'Other', options: []}
      Machine.find_each do |machine|
        if machine.has_machine_type?
          type = options.find {|c| c[:label] == machine.machine_type.category}
          if type.nil?
            type = {label: machine.machine_type.category, options: []}
            options << type
          end
          type[:options] << { label: machine.extended_name, value: target_to_id(machine), chart_type: target_to_chart_type(machine) }
        else
          other[:options] << { label: machine.extended_name, value: target_to_id(machine), chart_type: 'surfacing' }
        end
      end
      options << other
      options
    end

    def target_to_name(target)
      if target.is_a? Machine
        "#{target.name} (#{target.sn})"
      elsif target.is_a? MachineGroup
        target.name
      end
    end

    def target_to_id(target)
      if target.is_a? Machine
        "m-#{target.id}"
      elsif target.is_a? MachineGroup
        "g-#{target.id}"
      end
    end

    def target_to_chart_type(target)
      if target.is_a? Machine
        target.machine_type.category == 'box_coater' ? 'box_coater' : 'surfacing'
      elsif target.is_a? MachineGroup
        # surfacing: all machine in group has type != box_coater or type == nil
        return 'surfacing' if target.machines.all? { |g| !g.machine_type || g.machine_type.category != 'box_coater' }
        # box_coater: Machines in group have types that are mix of box coating and other machine types or nil
        return 'box_coater' if target.machines.all? { |m| m.machine_type && m.machine_type.category == 'box_coater' }
        # nil: All machines in group has machine type and machine type category is mix of box_coater and other types
        nil
      end
    end
  end
end
