module ReportComponents
  class MachinePerformanceOverview < ReportComponent
    include DataHelpers
    include ReportComponentHelpers
    include KpiHelpers

    class InvalidTimeFrame < StandardError; end

    def component_id
      'machine_performance_overview'
    end

    def component_type
      'bar-line'
    end

    def generate(args)
      # Because we are going to be stacking the utilization and uptime and that uptime contains utilization, we are going
      addons = [:as_percent]
      if use_planned_production_time
        addons << :planned_production_time
      end

      machine_states = MachineKpi.state_breakdown_between(start_time,
                                                          end_time,
                                                          timeframe,
                                                          machine_ids,
                                                          options: args.merge({ addons: addons })
      )

      waitingforjobs = extract_target_job_state(machine_states, :waitingforjobs)
      processing = extract_target_job_state(machine_states, :processing)

      throughput = MachineKpi.throughput_between(start_time, end_time, timeframe, machine_ids, options: args)

      data = {
        options: {
          metric: ProductionSetting.instance.throughput_metric.capitalize,
          targets: {
            uptime: args[:uptime_target],
            utilization: args[:utilization_target],
          }
        },
        series: [
          build_series_for_kpi_bucket('processing', processing, 'bar', :format_processing),
          build_series_for_kpi_bucket('waitingforjobs', waitingforjobs, 'bar', :format_waitingforjobs),
          build_series_for_kpi_bucket('throughput', throughput, 'line', :format_throughput)
        ]
      }

      extra_data = {}
      if use_planned_production_time && timeframe > Timeframe::HOUR_BY_HOUR
        extra_data[:active_time] = build_ppt_series(processing)
      end

      [data, extra_data]
    end

    def format_processing(bucket_value)
      bucket_value[:processing][:percent]
    end

    def format_waitingforjobs(bucket_value)
      bucket_value[:waitingforjobs][:percent]
    end

    def format_throughput(bucket_value)
      bucket_value[:count]
    end

    def build_ppt_series(result)
      result.map do |time, bucket_values|
        ppt_to_time_unit(bucket_values[:processing][:planned_production_time])
      end
    end

    def ppt_to_time_unit(planned_production_time)
      case timeframe
      when Timeframe::HOUR_BY_HOUR
        "#{(planned_production_time/60).round} minutes"
      when Timeframe::DAY_BY_DAY
        "#{(planned_production_time/60/60)} hours"
      when Timeframe::WEEK_BY_WEEK
        # neat round to quarter method
        rounded = ((planned_production_time/60/60/24) * 4.0).round / 4.0
        "#{rounded} days"
      when Timeframe::ALL
        "#{(planned_production_time/60/60)} hours"
      end
    end

    private

    # Should always be Timeframe::ALL so always one bucket?
    def extract_target_job_state(data, target_state)
      return data[target_state] if timeframe_all?

      new_data = {}
      data.each do |time_bucket, data|
        new_data[time_bucket] = data.slice(target_state)
      end
      new_data
    end
  end
end

