module ReportComponents
  class MachineProductivityTable < ReportComponent
    include DataHelpers
    include ReportComponentHelpers
    include KpiHelpers

    SINGLE_MACHINE_ARRAY_ID = 0
    MACHINE_GROUP_TARGET_ID = 1

    def component_id
      'machine_productivity_table'
    end

    def component_type
      'table'
    end

    def generate(args)
      # To optimize the calculations instead of doing machine kpi calculations for each target, we group the
      # single machine targets into a single machineKpi calculation without aggregation. Then we do machine groups
      # alone since those are already grouped. This means though that to keep the report row order we need to
      # keep an order array that maps where to get results.
      row_order_indexes = []
      targets_arr = [[],[]]

      args[:row_targets].each do |row_target|
        if row_target.is_a? MachineGroup
          targets_arr[MACHINE_GROUP_TARGET_ID] << row_target
          row_order_indexes << [MACHINE_GROUP_TARGET_ID, targets_arr[MACHINE_GROUP_TARGET_ID].length - 1]
        else
          targets_arr[SINGLE_MACHINE_ARRAY_ID] << row_target
          row_order_indexes << [SINGLE_MACHINE_ARRAY_ID, targets_arr[SINGLE_MACHINE_ARRAY_ID].length - 1]
        end
      end

      single_machine_uniq_ids = targets_arr[SINGLE_MACHINE_ARRAY_ID].map {|machine| machine.id}.uniq

      kpi_utilization = {}
      kpi_uptime = {}
      kpi_throughput = {}
      kpi_aborted = {}
      unless targets_arr[SINGLE_MACHINE_ARRAY_ID].empty?
        kpi_utilization = MachineKpi.utilization_between(start_time, end_time, timeframe, single_machine_uniq_ids, options: args.merge({addons: [:as_percent], aggregate_machines: false}))
        kpi_uptime = MachineKpi.uptime_between(start_time, end_time, timeframe, single_machine_uniq_ids, options: args.merge({addons: [:as_percent], aggregate_machines: false}))
        kpi_throughput = MachineKpi.throughput_between(start_time, end_time, timeframe, single_machine_uniq_ids, options: args.merge({aggregate_machines: false}))
        kpi_aborted = MachineKpi.aborted_job_ratio_between(start_time, end_time, timeframe, single_machine_uniq_ids, options: args.merge({aggregate_machines: false}))
      end

      series = []
      extra_data = []
      series_to_use_for_headers = nil

      row_order_indexes.each do |row_order_index|
        array_id, index = row_order_index
        target = targets_arr[array_id][index]
        if array_id == SINGLE_MACHINE_ARRAY_ID
          series.push(cell_series(utilization_row_title, kpi_utilization[target.id], :format_utilization, args))
          series.push(cell_series(uptime_row_title, kpi_uptime[target.id], :format_uptime, args))
          series.push(cell_series(throughput_row_title, kpi_throughput[target.id], :format_throughput, args))
          series.push(cell_series(aborts_row_title, kpi_aborted[target.id], :format_aborted, args))

          extra_data.push(target.extended_name)
          series_to_use_for_headers = kpi_utilization[target.id] if series_to_use_for_headers.nil?
        else
          # We only calculate for machine group on demand, do so now
          mg_utilization = MachineKpi.utilization_between(start_time, end_time, timeframe, target.machine_ids, options: args.merge({addons: [:as_percent], aggregate_machines: true}))
          mg_uptime = MachineKpi.uptime_between(start_time, end_time, timeframe, target.machine_ids, options: args.merge({addons: [:as_percent], aggregate_machines: true}))
          mg_throughput = MachineKpi.throughput_between(start_time, end_time, timeframe, target.machine_ids, options: args.merge({aggregate_machines: true}))
          mg_aborted = MachineKpi.aborted_job_ratio_between(start_time, end_time, timeframe, target.machine_ids, options: args.merge({aggregate_machines: true}))

          series.push(cell_series(utilization_row_title, mg_utilization, :format_utilization, args))
          series.push(cell_series(uptime_row_title, mg_uptime, :format_uptime, args))
          series.push(cell_series(throughput_row_title, mg_throughput, :format_throughput, args))
          series.push(cell_series(aborts_row_title, mg_aborted, :format_aborted, args))

          extra_data.push(target.name)
          series_to_use_for_headers = mg_utilization if series_to_use_for_headers.nil?
        end
      end

      [{
         headers: time_headers(series_to_use_for_headers),
         series: series
       }, {
        machine_names: extra_data
      }]
    end

    def multiple_separate_machines?(args)
      machine_ids.length > 0 && args[:expand_machine_group]
    end

    def format_utilization(bucket_value, args)
      {
        value: bucket_value[:percent].nil? ? nil : bucket_value[:percent].round.to_s,
        color: color_cell(bucket_value[:percent], args[:utilization_target], args)
      }
    end

    def format_uptime(bucket_value, args)
      {
        value: bucket_value[:percent].nil? ? nil : bucket_value[:percent].round.to_s,
        color: color_cell(bucket_value[:percent], args[:uptime_target], args)
      }
    end

    def format_throughput(bucket_value, args)
      {
        value: bucket_value[:count].nil? ? nil : bucket_value[:count].round.to_s,
        color: color_cell(bucket_value[:count], args[:throughput_target], args)
      }
    end

    def format_aborted(bucket_value, args)
      {
        value: bucket_value[:percent].nil? ? nil : bucket_value[:percent].round.to_s,
        color: color_cell(bucket_value[:percent], args[:aborted_target], args, true)
      }
    end

    def time_headers(single_series)
      single_series.map do |time, _|
        {
          value: bucket_time_to_timeframe_label(time)
        }
      end
    end

    def cell_series(name, single_series, formatter, args)
      [{value: name}] + single_series.map do |_, val|
        send(formatter, val, args)
      end
    end

    def bucket_time_to_timeframe_label(time)
      case timeframe
      when KpiHelpers::Timeframe::HOUR_BY_HOUR
        return "#{time.strftime('%H:%M')}"
      when KpiHelpers::Timeframe::DAY_BY_DAY
        return "#{time.strftime('%d/%m')}"
      when KpiHelpers::Timeframe::WEEK_BY_WEEK
        "#{time.strftime('%d/%m')} - #{(time + 1.week).strftime('%d/%m')}"
      end
    end

    def color_cell(value, target, args, invert=false)
      return grey if use_planned_production_time && value.nil?

      return white if multiple_separate_machines?(args)

      return low_is_good(value, target) if invert

      return white if target == nil
      return green if target == nil || target == 0
      return green if value >= target
      return yellow if value.to_f/target > 0.6
      red
    end

    def low_is_good(value, target)
      return white if target == nil
      return green if target == nil || target == 0
      return red if value >= target
      return yellow if value.to_f/target > 0.6
      green
    end

    def white
      'rgb(255, 255, 255)'
    end

    def green
      'rgb(0, 100, 0, .2)'
    end

    def yellow
      'rgb(231, 206, 115, .4)'
    end

    def red
      'rgb(150, 0, 0, .2)'
    end

    def grey
      'rgba(120, 120, 120, 0.5)'
    end

    def utilization_row_title
      'Utilization (%)'
    end

    def uptime_row_title
      'Uptime (%)'
    end

    def throughput_row_title
      "Throughput (#{ProductionSetting.instance.throughput_metric})"
    end

    def aborts_row_title
      'Aborts (%)'
    end
  end
end